# COPYRIGHT (C) 2020-2021 Nicotine+ Team
# COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
# COPYRIGHT (C) 2016 Mutnick <muhing@yahoo.com>
# COPYRIGHT (C) 2008-2011 Quinox <quinox@users.sf.net>
# COPYRIGHT (C) 2009 Daelstorm <daelstorm@gmail.com>
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import _thread

from ast import literal_eval
from time import time

from pynicotine import slskmessages
from pynicotine.logfacility import log


returncode = {
    'break': 0,  # don't give other plugins the event, do let n+ process it
    'zap': 1,    # don't give other plugins the event, don't let n+ process it
    'pass': 2    # do give other plugins the event, do let n+ process it
}                # returning nothing is the same as 'pass'


class PluginHandler(object):

    def __init__(self, frame, plugindir, config):
        self.frame = frame
        self.config = config

        log.add(_("Loading plugin handler"))

        self.my_username = self.config.sections["server"]["login"]
        self.plugindirs = []
        self.enabled_plugins = {}
        self.loaded_plugins = {}

        try:
            os.makedirs(plugindir)
        except Exception:
            pass

        # Load system-wide plugins
        prefix = os.path.dirname(os.path.realpath(__file__))
        self.plugindirs.append(os.path.join(prefix, "plugins"))

        # Load home directory plugins
        self.plugindirs.append(plugindir)

        if os.path.isdir(plugindir):
            self.load_enabled()
        else:
            log.add(_("It appears '%s' is not a directory, not loading plugins."), plugindir)

    def __findplugin(self, pluginname):
        for directory in self.plugindirs:
            fullpath = os.path.join(directory, pluginname)

            if os.path.exists(fullpath):
                return fullpath

        return None

    def toggle_plugin(self, pluginname):
        on = pluginname in self.enabled_plugins

        if on:
            _thread.start_new_thread(self.disable_plugin, (pluginname,))
        else:
            _thread.start_new_thread(self.enable_plugin, (pluginname,))

    def load_plugin(self, pluginname):

        try:
            # Import builtin plugin
            from importlib import import_module
            plugin = import_module("pynicotine.plugins." + pluginname)

        except Exception:
            # Import user plugin
            path = self.__findplugin(pluginname)

            if path is None:
                log.add(_("Failed to load plugin '%s', could not find it."), pluginname)
                return False

            from importlib.machinery import SourceFileLoader
            plugin = SourceFileLoader(pluginname, os.path.join(path, '__init__.py')).load_module()

        instance = plugin.Plugin(self)
        self.plugin_settings(pluginname, instance)
        instance.LoadNotification()

        self.loaded_plugins[pluginname] = plugin
        return plugin

    def enable_plugin(self, pluginname):

        # Our config file doesn't play nicely with some characters
        if "=" in pluginname:
            log.add(_("Unable to enable plugin %(name)s. Plugin folder name contains invalid characters: %(characters)s") % {"name": pluginname, "characters": "="})
            return

        if pluginname in self.enabled_plugins:
            return

        try:
            plugin = self.load_plugin(pluginname)

            if not plugin:
                raise Exception("Error loading plugin '%s'" % pluginname)

            plugin.enable(self)
            self.enabled_plugins[pluginname] = plugin
            log.add(_("Enabled plugin %s"), plugin.PLUGIN.__name__)

        except Exception:
            from traceback import print_exc
            print_exc()
            log.add(_("Unable to enable plugin %s"), pluginname)
            return False

        return True

    def list_installed_plugins(self):
        pluginlist = []

        for dir in self.plugindirs:
            if os.path.exists(dir):
                for file in os.listdir(dir):
                    if file not in pluginlist and os.path.isdir(os.path.join(dir, file)):
                        pluginlist.append(file)

        return pluginlist

    def disable_plugin(self, pluginname):
        if pluginname not in self.enabled_plugins:
            return

        try:
            plugin = self.enabled_plugins[pluginname]

            log.add(_("Disabled plugin {}".format(plugin.PLUGIN.__name__)))
            del self.enabled_plugins[pluginname]
            plugin.disable(self)

        except Exception:
            from traceback import print_exc
            print_exc()
            log.add(_("Unable to fully disable plugin %s"), pluginname)
            return False

        return True

    def get_plugin_settings(self, pluginname):
        if pluginname in self.enabled_plugins:
            plugin = self.enabled_plugins[pluginname]

            if hasattr(plugin.PLUGIN, "metasettings"):
                return plugin.PLUGIN.metasettings

    def get_plugin_info(self, pluginname):
        path = os.path.join(self.__findplugin(pluginname), 'PLUGININFO')

        with open(path) as f:
            infodict = {}

            for line in f:
                try:
                    key, val = line.split("=", 1)
                    infodict[key.strip()] = literal_eval(val.strip())
                except ValueError:
                    pass  # this happens on blank lines

        return infodict

    def save_enabled(self):
        self.config.sections["plugins"]["enabled"] = list(self.enabled_plugins.keys())

    def load_enabled(self):
        enable = self.config.sections["plugins"]["enable"]

        if not enable:
            return

        to_enable = self.config.sections["plugins"]["enabled"]

        for plugin in to_enable:
            _thread.start_new_thread(self.enable_plugin, (plugin,))

    def plugin_settings(self, pluginname, plugin):
        try:
            if not hasattr(plugin, "settings"):
                return

            if pluginname not in self.config.sections["plugins"]:
                self.config.sections["plugins"][pluginname] = plugin.settings

            for i in plugin.settings:
                if i not in self.config.sections["plugins"][pluginname]:
                    self.config.sections["plugins"][pluginname][i] = plugin.settings[i]

            customsettings = self.config.sections["plugins"][pluginname]

            for key in customsettings:
                if key in plugin.settings:
                    plugin.settings[key] = customsettings[key]

                else:
                    log.add(_("Stored setting '%(name)s' is no longer present in the plugin"), {'name': key})

        except KeyError:
            log.add("No custom settings found for %s", (plugin.__name__,))

    def trigger_public_command_event(self, room, command, args):
        return self._trigger_command(command, room, args, public_command=True)

    def trigger_private_command_event(self, user, command, args):
        return self._trigger_command(command, user, args, public_command=False)

    def _trigger_command(self, command, source, args, public_command):
        for module, plugin in self.enabled_plugins.items():
            try:
                if plugin.PLUGIN is None:
                    continue

                if public_command:
                    ret = plugin.PLUGIN.PublicCommandEvent(command, source, args)
                else:
                    ret = plugin.PLUGIN.PrivateCommandEvent(command, source, args)

                if ret is not None:
                    if ret == returncode['zap']:
                        return True
                    elif ret == returncode['pass']:
                        pass
                    else:
                        log.add(_("Plugin %(module)s returned something weird, '%(value)s', ignoring"), {'module': module, 'value': str(ret)})

            except Exception:
                from traceback import extract_stack
                from traceback import extract_tb
                from traceback import format_list

                log.add(_("Plugin %(module)s failed with error %(errortype)s: %(error)s.\nTrace: %(trace)s\nProblem area:%(area)s"), {
                    'module': module,
                    'errortype': sys.exc_info()[0],
                    'error': sys.exc_info()[1],
                    'trace': ''.join(format_list(extract_stack())),
                    'area': ''.join(format_list(extract_tb(sys.exc_info()[2])))
                })

        return False

    def trigger_event(self, function, args):
        """Triggers an event for the plugins. Since events and notifications
        are precisely the same except for how n+ responds to them, both can be
        triggered by this function."""

        hotpotato = args

        for module, plugin in self.enabled_plugins.items():
            try:
                ret = getattr(plugin.PLUGIN, function)(*hotpotato)

                if ret is not None and not isinstance(ret, tuple):
                    if ret == returncode['zap']:
                        return None
                    elif ret == returncode['break']:
                        return hotpotato
                    elif ret == returncode['pass']:
                        pass
                    else:
                        log.add(_("Plugin %(module)s returned something weird, '%(value)s', ignoring"), {'module': module, 'value': ret})

                if ret is not None:
                    hotpotato = ret

            except Exception:
                from traceback import extract_stack
                from traceback import extract_tb
                from traceback import format_list

                log.add(_("Plugin %(module)s failed with error %(errortype)s: %(error)s.\nTrace: %(trace)s\nProblem area:%(area)s"), {
                    'module': module,
                    'errortype': sys.exc_info()[0],
                    'error': sys.exc_info()[1],
                    'trace': ''.join(format_list(extract_stack())),
                    'area': ''.join(format_list(extract_tb(sys.exc_info()[2])))
                })

        return hotpotato

    def search_request_notification(self, searchterm, user, searchid):
        self.trigger_event("SearchRequestNotification", (searchterm, user, searchid))

    def distrib_search_notification(self, searchterm, user, searchid):
        self.trigger_event("DistribSearchNotification", (searchterm, user, searchid))

    def public_room_message_notification(self, room, user, line):
        self.trigger_event("PublicRoomMessageNotification", (room, user, line))

    def incoming_private_chat_event(self, user, line):
        if user != self.my_username:
            # dont trigger the scripts on our own talking - we've got "Outgoing" for that
            return self.trigger_event("IncomingPrivateChatEvent", (user, line))
        else:
            return (user, line)

    def incoming_private_chat_notification(self, user, line):
        self.trigger_event("IncomingPrivateChatNotification", (user, line))

    def incoming_public_chat_event(self, room, user, line):
        return self.trigger_event("IncomingPublicChatEvent", (room, user, line))

    def incoming_public_chat_notification(self, room, user, line):
        self.trigger_event("IncomingPublicChatNotification", (room, user, line))

    def outgoing_private_chat_event(self, user, line):
        if line is not None:
            # if line is None nobody actually said anything
            return self.trigger_event("OutgoingPrivateChatEvent", (user, line))
        else:
            return (user, line)

    def outgoing_private_chat_notification(self, user, line):
        self.trigger_event("OutgoingPrivateChatNotification", (user, line))

    def outgoing_public_chat_event(self, room, line):
        return self.trigger_event("OutgoingPublicChatEvent", (room, line))

    def outgoing_public_chat_notification(self, room, line):
        self.trigger_event("OutgoingPublicChatNotification", (room, line))

    def outgoing_global_search_event(self, text):
        return self.trigger_event("OutgoingGlobalSearchEvent", (text,))

    def outgoing_room_search_event(self, rooms, text):
        return self.trigger_event("OutgoingRoomSearchEvent", (rooms, text))

    def outgoing_buddy_search_event(self, text):
        return self.trigger_event("OutgoingBuddySearchEvent", (text,))

    def outgoing_user_search_event(self, users, text):
        return self.trigger_event("OutgoingUserSearchEvent", (users, text))

    def user_resolve_notification(self, user, ip, port, country=None):
        """Notification for user IP:Port resolving.

        Note that country is only set when the user requested the resolving"""
        self.trigger_event("UserResolveNotification", (user, ip, port, country))

    def server_connect_notification(self):
        self.trigger_event("ServerConnectNotification", (),)

    def server_disconnect_notification(self, userchoice):
        self.trigger_event("ServerDisconnectNotification", (userchoice, ))

    def join_chatroom_notification(self, room):
        self.trigger_event("JoinChatroomNotification", (room,))

    def leave_chatroom_notification(self, room):
        self.trigger_event("LeaveChatroomNotification", (room,))

    def upload_queued_notification(self, user, virtualfile, realfile):
        self.trigger_event("UploadQueuedNotification", (user, virtualfile, realfile))

    def user_stats_notification(self, user, stats):
        self.trigger_event("UserStatsNotification", (user, stats))

    """ Other Functions """

    def log(self, text):
        log.add(text)

    def saychatroom(self, room, text):
        self.frame.np.queue.put(slskmessages.SayChatroom(room, text))

    def sayprivate(self, user, text):
        '''Send user message in private (showing up in GUI)'''
        self.frame.privatechats.users[user].send_message(text)

    def sendprivate(self, user, text):
        '''Send user message in private (not showing up in GUI)'''
        self.frame.privatechats.send_message(user, text)


class ResponseThrottle(object):

    """
    ResponseThrottle - Mutnick 2016

    See 'reddit' and my other plugins for example use

    Purpose: Avoid flooding chat room with plugin responses
        Some plugins respond based on user requests and we do not want
        to respond too much and encounter a temporary server chat ban

    Some of the throttle logic is guesswork as server code is closed source, but works adequately.
    """

    def __init__(self, frame, plugin_name, logging=False):
        self.frame = frame
        self.plugin_name = plugin_name
        self.logging = logging
        self.plugin_usage = {}

    def ok_to_respond(self, room, nick, request, seconds_limit_min=30):
        self.room = room
        self.nick = nick
        self.request = request

        willing_to_respond = True
        current_time = time()

        if room not in self.plugin_usage:
            self.plugin_usage[room] = {'last_time': 0, 'last_request': "", 'last_nick': ""}

        last_time = self.plugin_usage[room]['last_time']
        last_nick = self.plugin_usage[room]['last_nick']
        last_request = self.plugin_usage[room]['last_request']

        port = False
        try:
            ip, port = self.frame.np.users[nick].addr
        except Exception:
            port = True

        if nick in self.frame.np.config.sections["server"]["ignorelist"]:
            willing_to_respond, reason = False, "The nick is ignored"

        elif self.frame.user_ip_is_ignored(nick):
            willing_to_respond, reason = False, "The nick's Ip is ignored"

        elif not port:
            willing_to_respond, reason = False, "Request likely from simple PHP based griefer bot"

        elif [nick, request] == [last_nick, last_request]:
            if (current_time - last_time) < 12 * seconds_limit_min:
                willing_to_respond, reason = False, "Too soon for same nick to request same resource in room"

        elif (request == last_request):
            if (current_time - last_time) < 3 * seconds_limit_min:
                willing_to_respond, reason = False, "Too soon for different nick to request same resource in room"

        else:
            recent_responses = 0

            for responded_room in self.plugin_usage:
                if (current_time - self.plugin_usage[responded_room]['last_time']) < seconds_limit_min:
                    recent_responses += 1

                    if responded_room == room:
                        willing_to_respond, reason = False, "Responded in specified room too recently"
                        break

            if recent_responses > 3:
                willing_to_respond, reason = False, "Responded in multiple rooms enough"

        if self.logging:
            if not willing_to_respond:
                base_log_msg = "{} plugin request rejected - room '{}', nick '{}'".format(self.plugin_name, room, nick)
                log.add_debug("{} - {}".format(base_log_msg, reason))

        return willing_to_respond

    def responded(self, msg=""):
        # possible TODO's: we could actually say public the msg here
        # make more stateful - track past msg's as additional responder willingness criteria, etc
        self.plugin_usage[self.room] = {'last_time': time(), 'last_request': self.request, 'last_nick': self.nick}


class BasePlugin(object):

    __name__ = "BasePlugin"
    __desc__ = "No description provided"
    __version__ = "2016-08-30"
    __publiccommands__ = []
    __privatecommands__ = []

    def __init__(self, parent):
        # Never override this function, override init() instead
        self.parent = parent
        self.frame = parent.frame

        self.init()
        for (trigger, func) in self.__publiccommands__:
            self.frame.chatrooms.CMDS.add('/' + trigger + ' ')
        for (trigger, func) in self.__privatecommands__:
            self.frame.privatechats.CMDS.add('/' + trigger + ' ')

    def init(self):
        # Custom init function for plugins
        pass

    def LoadSettings(self, settings):  # noqa
        self.settings = settings

    def LoadNotification(self):  # noqa
        pass

    def PublicRoomMessageNotification(self, room, user, line):  # noqa
        pass

    def SearchRequestNotification(self, searchterm, user, searchid):  # noqa
        pass

    def DistribSearchNotification(self, searchterm, user, searchid):  # noqa
        pass

    def IncomingPrivateChatEvent(self, user, line):  # noqa
        pass

    def IncomingPrivateChatNotification(self, user, line):  # noqa
        pass

    def IncomingPublicChatEvent(self, room, user, line):  # noqa
        pass

    def IncomingPublicChatNotification(self, room, user, line):  # noqa
        pass

    def OutgoingPrivateChatEvent(self, user, line):  # noqa
        pass

    def OutgoingPrivateChatNotification(self, user, line):  # noqa
        pass

    def OutgoingPublicChatEvent(self, room, line):  # noqa
        pass

    def OutgoingPublicChatNotification(self, room, line):  # noqa
        pass

    def OutgoingGlobalSearchEvent(self, text):  # noqa
        pass

    def OutgoingRoomSearchEvent(self, rooms, text):  # noqa
        pass

    def OutgoingBuddySearchEvent(self, text):  # noqa
        pass

    def OutgoingUserSearchEvent(self, users, text):  # noqa
        pass

    def UserResolveNotification(self, user, ip, port, country):  # noqa
        pass

    def ServerConnectNotification(self):  # noqa
        pass

    def ServerDisconnectNotification(self, userchoice):  # noqa
        pass

    def JoinChatroomNotification(self, room):  # noqa
        pass

    def LeaveChatroomNotification(self, room):  # noqa
        pass

    def UploadQueuedNotification(self, user, virtualfile, realfile):  # noqa
        pass

    def UserStatsNotification(self, user, stats):  # noqa
        pass

    # The following are functions to make your life easier,
    # you shouldn't override them.
    def log(self, text):
        self.parent.log(self.__name__ + ": " + text)

    def saypublic(self, room, text):
        self.parent.saychatroom(room, text)

    def sayprivate(self, user, text):
        '''Send user message in private (shows up in GUI)'''
        self.parent.sayprivate(user, text)

    def sendprivate(self, user, text):
        '''Send user message in private (doesn't show up in GUI)'''
        self.parent.sendprivate(user, text)

    def fakepublic(self, room, user, text):
        try:
            room = self.frame.chatrooms.joinedrooms[room]
        except KeyError:
            return False

        msg = slskmessages.SayChatroom(room, text)
        msg.user = user
        room.say_chat_room(msg, text)
        return True

    # The following are functions used by the plugin system,
    # you are not allowed to override these.
    def PublicCommandEvent(self, command, room, args):  # noqa
        for (trigger, func) in self.__publiccommands__:
            if trigger == command:
                return func(self, room, args)

    def PrivateCommandEvent(self, command, user, args):  # noqa
        for (trigger, func) in self.__privatecommands__:
            if trigger == command:
                return func(self, user, args)



# COPYRIGHT (C) 2020-2021 Nicotine+ Team
# COPYRIGHT (C) 2020 Lene Preuss <lene.preuss@gmail.com>
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct

from pynicotine.slskmessages import FileSearch
from pynicotine.slskproto import SlskProtoThread

# "Magic" values confirmed to work with nicotine+ 1.4.2
SEARCH_ID = 524700074
SEARCH_TEXT = '70 gwen auto'
SEARCH_NETWORK_VALUE = [
    170, 73, 70, 31, 12, 0, 0, 0,  # search id
    55, 48, 32, 103, 119, 101, 110, 32, 97, 117, 116, 111  # search text
]
SEARCH_OUT_MSG = [
    24, 0, 0, 0,  # length of message starting with next byte
    26, 0, 0, 0,  # server code for FileSearch
    170, 73, 70, 31, 12, 0, 0, 0,
    55, 48, 32, 103, 119, 101, 110, 32, 97, 117, 116, 111
]


def test_instantiate_file_search():
    search = FileSearch(requestid=SEARCH_ID, text=SEARCH_TEXT)
    assert search.searchid == SEARCH_ID
    assert search.searchterm == SEARCH_TEXT


def test_make_network_message():
    search = FileSearch(requestid=SEARCH_ID, text=SEARCH_TEXT)
    assert [b for b in search.make_network_message()] == SEARCH_NETWORK_VALUE


def test_sent_out_message():
    search = FileSearch(requestid=SEARCH_ID, text=SEARCH_TEXT)
    msg = search.make_network_message()
    out_msg = struct.pack("<ii", len(msg) + 4, SlskProtoThread.servercodes[search.__class__]) + msg
    assert [b for b in out_msg] == SEARCH_OUT_MSG



"""
MIT License

Copyright (c) 2017 IP2Location.com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import struct
import socket

if sys.version < '3':
    def u(x):
        return x.decode('utf-8')

    def b(x):
        return str(x)
else:
    def u(x):
        if isinstance(x, bytes):
            return x.decode()
        return x

    def b(x):
        if isinstance(x, bytes):
            return x
        return x.encode('ascii')

# Windows version of Python does not provide it
#          for compatibility with older versions of Windows.
if not hasattr(socket, 'inet_pton'):
    def inet_pton(t, addr):
        import ctypes
        a = ctypes.WinDLL('ws2_32.dll')
        in_addr_p = ctypes.create_string_buffer(b(addr))
        if t == socket.AF_INET:
            out_addr_p = ctypes.create_string_buffer(4)
        elif t == socket.AF_INET6:
            out_addr_p = ctypes.create_string_buffer(16)
        n = a.inet_pton(t, in_addr_p, out_addr_p)
        if n == 0:
            raise ValueError('Invalid address')
        return out_addr_p.raw
    socket.inet_pton = inet_pton


class IP2LocationRecord:
    ''' IP2Location record with all fields from the database '''
    ip = None
    country_short = None
    country_long = None
    region = None
    city = None
    isp = None
    latitude = None
    longitude = None
    domain = None
    zipcode = None
    timezone = None
    netspeed = None
    idd_code = None
    area_code = None
    weather_code = None
    weather_name = None
    mcc = None
    mnc = None
    mobile_brand = None
    elevation = None
    usage_type = None

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return repr(self.__dict__)


max_ipv4_range = 4294967295
max_ipv6_range = 340282366920938463463374607431768211455

_COUNTRY_POSITION = (0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2)
_REGION_POSITION = (0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3)
_CITY_POSITION = (0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4)
_ISP_POSITION = (0, 0, 3, 0, 5, 0, 7, 5, 7, 0, 8, 0, 9, 0, 9, 0, 9, 0, 9, 7, 9, 0, 9, 7, 9)
_LATITUDE_POSITION = (0, 0, 0, 0, 0, 5, 5, 0, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5)
_LONGITUDE_POSITION = (0, 0, 0, 0, 0, 6, 6, 0, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6)
_DOMAIN_POSITION = (0, 0, 0, 0, 0, 0, 0, 6, 8, 0, 9, 0, 10, 0, 10, 0, 10, 0, 10, 8, 10, 0, 10, 8, 10)
_ZIPCODE_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 7, 7, 7, 0, 7, 7, 7, 0, 7, 0, 7, 7, 7, 0, 7)
_TIMEZONE_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, 8, 7, 8, 8, 8, 7, 8, 0, 8, 8, 8, 0, 8)
_NETSPEED_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, 11, 0, 11, 8, 11, 0, 11, 0, 11, 0, 11)
_IDDCODE_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 12, 0, 12, 0, 12, 9, 12, 0, 12)
_AREACODE_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 13, 0, 13, 0, 13, 10, 13, 0, 13)
_WEATHERSTATIONCODE_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 14, 0, 14, 0, 14, 0, 14)
_WEATHERSTATIONNAME_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 15, 0, 15, 0, 15, 0, 15)
_MCC_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 16, 0, 16, 9, 16)
_MNC_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 17, 0, 17, 10, 17)
_MOBILEBRAND_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11, 18, 0, 18, 11, 18)
_ELEVATION_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11, 19, 0, 19)
_USAGETYPE_POSITION = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 20)


class IP2Location(object):
    ''' IP2Location database '''

    def __init__(self, filename=None, mode='FILE_IO'):
        ''' Creates a database object and opens a file if filename is given

        '''
        self.mode = mode
        if filename:
            self.open(filename)

    def __enter__(self):
        if not hasattr(self, '_f') or self._f.closed:
            raise ValueError("Cannot enter context with closed file")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def open(self, filename):
        ''' Opens a database file '''
        # Ensure old file is closed before opening a new one
        self.close()

        if (self.mode == 'SHARED_MEMORY'):
            import mmap
            db1 = open(filename, 'rb')
            try:
                self._f = mmap.mmap(db1.fileno(), 0, prot=mmap.PROT_READ)
            except AttributeError:
                self._f = mmap.mmap(db1.fileno(), 0, access=mmap.ACCESS_READ)
            db1.close()
            del db1
        elif (self.mode == 'FILE_IO'):
            self._f = open(filename, 'rb')
        else:
            raise ValueError("Invalid mode. Please enter either FILE_IO or SHARED_MEMORY.")
        self._dbtype = struct.unpack('B', self._f.read(1))[0]
        self._dbcolumn = struct.unpack('B', self._f.read(1))[0]
        self._dbyear = struct.unpack('B', self._f.read(1))[0]
        self._dbmonth = struct.unpack('B', self._f.read(1))[0]
        self._dbday = struct.unpack('B', self._f.read(1))[0]
        self._ipv4dbcount = struct.unpack('<I', self._f.read(4))[0]
        self._ipv4dbaddr = struct.unpack('<I', self._f.read(4))[0]
        self._ipv6dbcount = struct.unpack('<I', self._f.read(4))[0]
        self._ipv6dbaddr = struct.unpack('<I', self._f.read(4))[0]
        self._ipv4indexbaseaddr = struct.unpack('<I', self._f.read(4))[0]
        self._ipv6indexbaseaddr = struct.unpack('<I', self._f.read(4))[0]

    def close(self):
        if hasattr(self, '_f'):
            # If there is file close it.
            self._f.close()
            del self._f

    def get_country_short(self, ip):
        ''' Get country_short '''
        rec = self.get_all(ip)
        return rec and rec.country_short

    def get_country_long(self, ip):
        ''' Get country_long '''
        rec = self.get_all(ip)
        return rec and rec.country_long

    def get_region(self, ip):
        ''' Get region '''
        rec = self.get_all(ip)
        return rec and rec.region

    def get_city(self, ip):
        ''' Get city '''
        rec = self.get_all(ip)
        return rec and rec.city

    def get_isp(self, ip):
        ''' Get isp '''
        rec = self.get_all(ip)
        return rec and rec.isp

    def get_latitude(self, ip):
        ''' Get latitude '''
        rec = self.get_all(ip)
        return rec and rec.latitude

    def get_longitude(self, ip):
        ''' Get longitude '''
        rec = self.get_all(ip)
        return rec and rec.longitude

    def get_domain(self, ip):
        ''' Get domain '''
        rec = self.get_all(ip)
        return rec and rec.domain

    def get_zipcode(self, ip):
        ''' Get zipcode '''
        rec = self.get_all(ip)
        return rec and rec.zipcode

    def get_timezone(self, ip):
        ''' Get timezone '''
        rec = self.get_all(ip)
        return rec and rec.timezone

    def get_netspeed(self, ip):
        ''' Get netspeed '''
        rec = self.get_all(ip)
        return rec and rec.netspeed

    def get_idd_code(self, ip):
        ''' Get idd_code '''
        rec = self.get_all(ip)
        return rec and rec.idd_code

    def get_area_code(self, ip):
        ''' Get area_code '''
        rec = self.get_all(ip)
        return rec and rec.area_code

    def get_weather_code(self, ip):
        ''' Get weather_code '''
        rec = self.get_all(ip)
        return rec and rec.weather_code

    def get_weather_name(self, ip):
        ''' Get weather_name '''
        rec = self.get_all(ip)
        return rec and rec.weather_name

    def get_mcc(self, ip):
        ''' Get mcc '''
        rec = self.get_all(ip)
        return rec and rec.mcc

    def get_mnc(self, ip):
        ''' Get mnc '''
        rec = self.get_all(ip)
        return rec and rec.mnc

    def get_mobile_brand(self, ip):
        ''' Get mobile_brand '''
        rec = self.get_all(ip)
        return rec and rec.mobile_brand

    def get_elevation(self, ip):
        ''' Get elevation '''
        rec = self.get_all(ip)
        return rec and rec.elevation

    def get_usage_type(self, ip):
        ''' Get usage_type '''
        rec = self.get_all(ip)
        return rec and rec.usage_type

    def get_all(self, addr):
        ''' Get the whole record with all fields read from the file

            Arguments:

            addr: IPv4 or IPv6 address as a string

            Returns IP2LocationRecord or None if address not found in file
        '''
        return self._get_record(addr)

    def find(self, addr):
        ''' Get the whole record with all fields read from the file

            Arguments:

            addr: IPv4 or IPv6 address as a string

            Returns IP2LocationRecord or None if address not found in file
        '''
        return self._get_record(addr)

    def _reads(self, offset):
        self._f.seek(offset - 1)
        n = struct.unpack('B', self._f.read(1))[0]

        if sys.version < '3':
            return str(self._f.read(n).decode('iso-8859-1').encode('utf-8'))
        else:
            return u(self._f.read(n).decode('iso-8859-1').encode('utf-8'))

    def _readi(self, offset):
        self._f.seek(offset - 1)
        return struct.unpack('<I', self._f.read(4))[0]

    def _readf(self, offset):
        self._f.seek(offset - 1)
        return struct.unpack('<f', self._f.read(4))[0]

    def _readip(self, offset, ipv):
        if ipv == 4:
            return self._readi(offset)
        elif ipv == 6:
            a, b, c, d = self._readi(offset), self._readi(offset + 4), self._readi(offset + 8), self._readi(offset + 12)
            return (d << 96) | (c << 64) | (b << 32) | a

    def _readips(self, offset, ipv):
        if ipv == 4:
            return socket.inet_ntoa(struct.pack('!L', self._readi(offset)))
        elif ipv == 6:
            return str(self._readip(offset, ipv))

    def _read_record(self, mid, ipv):
        rec = IP2LocationRecord()

        if ipv == 4:
            off = 0
            baseaddr = self._ipv4dbaddr
            dbcolumn_width = self._dbcolumn * 4 + 4
        elif ipv == 6:
            off = 12
            baseaddr = self._ipv6dbaddr
            dbcolumn_width = self._dbcolumn * 4

        def calc_off(what, mid):
            return baseaddr + mid * (self._dbcolumn * 4 + off) + off + 4 * (what[self._dbtype] - 1)

        if (self.mode == 'SHARED_MEMORY'):
            # We can directly use slice notation to read content from mmap object. https://docs.python.org/3/library/mmap.html?highlight=mmap#module-mmap
            raw_positions_row = self._f[(calc_off(_COUNTRY_POSITION, mid)) - 1: (calc_off(_COUNTRY_POSITION, mid)) - 1 + dbcolumn_width]
        else:
            self._f.seek((calc_off(_COUNTRY_POSITION, mid)) - 1)
            raw_positions_row = self._f.read(dbcolumn_width)

        if self.original_ip != '':
            rec.ip = self.original_ip
        else:
            rec.ip = self._readips(baseaddr + (mid) * self._dbcolumn * 4, ipv)

        if _COUNTRY_POSITION[self._dbtype] != 0:
            rec.country_short = self._reads(struct.unpack('<I', raw_positions_row[0: ((_COUNTRY_POSITION[self._dbtype] - 1) * 4)])[0] + 1)
            rec.country_long = self._reads(struct.unpack('<I', raw_positions_row[0: ((_COUNTRY_POSITION[self._dbtype] - 1) * 4)])[0] + 4)

        if _REGION_POSITION[self._dbtype] != 0:
            rec.region = self._reads(struct.unpack('<I', raw_positions_row[((_REGION_POSITION[self._dbtype] - 1) * 4 - 4): ((_REGION_POSITION[self._dbtype] - 1) * 4)])[0] + 1)
        if _CITY_POSITION[self._dbtype] != 0:
            rec.city = self._reads(struct.unpack('<I', raw_positions_row[((_CITY_POSITION[self._dbtype] - 1) * 4 - 4): ((_CITY_POSITION[self._dbtype] - 1) * 4)])[0] + 1)
        if _ISP_POSITION[self._dbtype] != 0:
            rec.isp = self._reads(struct.unpack('<I', raw_positions_row[((_ISP_POSITION[self._dbtype] - 1) * 4 - 4): ((_ISP_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _LATITUDE_POSITION[self._dbtype] != 0:
            rec.latitude = round(struct.unpack('<f', raw_positions_row[((_LATITUDE_POSITION[self._dbtype] - 1) * 4 - 4): ((_LATITUDE_POSITION[self._dbtype] - 1) * 4)])[0], 6)
        if _LONGITUDE_POSITION[self._dbtype] != 0:
            rec.longitude = round(struct.unpack('<f', raw_positions_row[((_LONGITUDE_POSITION[self._dbtype] - 1) * 4 - 4): ((_LONGITUDE_POSITION[self._dbtype] - 1) * 4)])[0], 6)

        if _DOMAIN_POSITION[self._dbtype] != 0:
            rec.domain = self._reads(struct.unpack('<I', raw_positions_row[((_DOMAIN_POSITION[self._dbtype] - 1) * 4 - 4): ((_DOMAIN_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _ZIPCODE_POSITION[self._dbtype] != 0:
            rec.zipcode = self._reads(struct.unpack('<I', raw_positions_row[((_ZIPCODE_POSITION[self._dbtype] - 1) * 4 - 4): ((_ZIPCODE_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _TIMEZONE_POSITION[self._dbtype] != 0:
            rec.timezone = self._reads(struct.unpack('<I', raw_positions_row[((_TIMEZONE_POSITION[self._dbtype] - 1) * 4 - 4): ((_TIMEZONE_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _NETSPEED_POSITION[self._dbtype] != 0:
            rec.netspeed = self._reads(struct.unpack('<I', raw_positions_row[((_NETSPEED_POSITION[self._dbtype] - 1) * 4 - 4): ((_NETSPEED_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _IDDCODE_POSITION[self._dbtype] != 0:
            rec.idd_code = self._reads(struct.unpack('<I', raw_positions_row[((_IDDCODE_POSITION[self._dbtype] - 1) * 4 - 4): ((_IDDCODE_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _AREACODE_POSITION[self._dbtype] != 0:
            rec.area_code = self._reads(struct.unpack('<I', raw_positions_row[((_AREACODE_POSITION[self._dbtype] - 1) * 4 - 4): ((_AREACODE_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _WEATHERSTATIONCODE_POSITION[self._dbtype] != 0:
            rec.weather_code = self._reads(struct.unpack('<I', raw_positions_row[((_WEATHERSTATIONCODE_POSITION[self._dbtype] - 1) * 4 - 4): ((_WEATHERSTATIONCODE_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _WEATHERSTATIONNAME_POSITION[self._dbtype] != 0:
            rec.weather_name = self._reads(struct.unpack('<I', raw_positions_row[((_WEATHERSTATIONNAME_POSITION[self._dbtype] - 1) * 4 - 4): ((_WEATHERSTATIONNAME_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _MCC_POSITION[self._dbtype] != 0:
            rec.mcc = self._reads(struct.unpack('<I', raw_positions_row[((_MCC_POSITION[self._dbtype] - 1) * 4 - 4): ((_MCC_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _MNC_POSITION[self._dbtype] != 0:
            rec.mnc = self._reads(struct.unpack('<I', raw_positions_row[((_MNC_POSITION[self._dbtype] - 1) * 4 - 4): ((_MNC_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _MOBILEBRAND_POSITION[self._dbtype] != 0:
            rec.mobile_brand = self._reads(struct.unpack('<I', raw_positions_row[((_MOBILEBRAND_POSITION[self._dbtype] - 1) * 4 - 4): ((_MOBILEBRAND_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _ELEVATION_POSITION[self._dbtype] != 0:
            rec.elevation = self._reads(struct.unpack('<I', raw_positions_row[((_ELEVATION_POSITION[self._dbtype] - 1) * 4 - 4): ((_ELEVATION_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        if _USAGETYPE_POSITION[self._dbtype] != 0:
            rec.usage_type = self._reads(struct.unpack('<I', raw_positions_row[((_USAGETYPE_POSITION[self._dbtype] - 1) * 4 - 4): ((_USAGETYPE_POSITION[self._dbtype] - 1) * 4)])[0] + 1)

        return rec

    def __iter__(self):
        low, high = 0, self._ipv4dbcount
        while low <= high:
            yield self._read_record(low, 4)
            low += 1

        low, high = 0, self._ipv6dbcount
        while low <= high:
            yield self._read_record(low, 6)
            low += 1

    def _parse_addr(self, addr):
        ''' Parses address and returns IP version. Raises exception on invalid argument '''
        ipv = 0
        try:
            a, b = struct.unpack('!QQ', socket.inet_pton(socket.AF_INET6, addr))
            ipnum = (a << 64) | b
            # Convert ::FFFF:x.y.z.y to IPv4
            if addr.lower().startswith('::ffff:'):
                try:
                    socket.inet_pton(socket.AF_INET, addr)
                    ipv = 4
                except Exception:
                    # reformat ipv4 address in ipv6
                    if ((ipnum >= 281470681743360) and (ipnum <= 281474976710655)):
                        ipv = 4
                        ipnum = ipnum - 281470681743360
                    else:
                        ipv = 6
            else:
                # reformat 6to4 address to ipv4 address 2002:: to 2002:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF
                if ((ipnum >= 42545680458834377588178886921629466624) and (ipnum <= 42550872755692912415807417417958686719)):
                    ipv = 4
                    ipnum = ipnum >> 80
                    ipnum = ipnum % 4294967296
                # reformat Teredo address to ipv4 address 2001:0000:: to 2001:0000:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:
                elif ((ipnum >= 42540488161975842760550356425300246528) and (ipnum <= 42540488241204005274814694018844196863)):
                    ipv = 4
                    ipnum = ~ ipnum
                    ipnum = ipnum % 4294967296
                else:
                    ipv = 6
        except Exception:
            ipnum = struct.unpack('!L', socket.inet_pton(socket.AF_INET, addr))[0]
            ipv = 4
        return ipv, ipnum

    def _get_record(self, ip):

        self.original_ip = ip
        low = 0
        ipv = self._parse_addr(ip)[0]
        ipnum = self._parse_addr(ip)[1]
        if ipv == 4:
            if (ipnum == max_ipv4_range):
                ipno = ipnum - 1
            else:
                ipno = ipnum
            off = 0
            baseaddr = self._ipv4dbaddr
            high = self._ipv4dbcount
            if self._ipv4indexbaseaddr > 0:
                indexpos = ((ipno >> 16) << 3) + self._ipv4indexbaseaddr
                low = self._readi(indexpos)
                high = self._readi(indexpos + 4)

        elif ipv == 6:
            if self._ipv6dbcount == 0:
                raise ValueError('Please use IPv6 BIN file for IPv6 Address.')

            if (ipnum == max_ipv6_range):
                ipno = ipnum - 1
            else:
                ipno = ipnum
            off = 12
            baseaddr = self._ipv6dbaddr
            high = self._ipv6dbcount
            if self._ipv6indexbaseaddr > 0:
                indexpos = ((ipno >> 112) << 3) + self._ipv6indexbaseaddr
                low = self._readi(indexpos)
                high = self._readi(indexpos + 4)

        while low <= high:
            mid = int((low + high) / 2)
            ipfrom = self._readip(baseaddr + (mid) * (self._dbcolumn * 4 + off), ipv)
            ipto = self._readip(baseaddr + (mid + 1) * (self._dbcolumn * 4 + off), ipv)

            if ipfrom <= ipno < ipto:
                return self._read_record(mid, ipv)
            else:
                if ipno < ipfrom:
                    high = mid - 1
                else:
                    low = mid + 1



# COPYRIGHT (C) 2020-2021 Nicotine+ Team
# COPYRIGHT (C) 2009 quinox. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


def code2name(code):
    try:
        return lookup[code.lower()]

    except KeyError:
        return code


lookup = {
    'ad': _('Andorra'),
    'ae': _('United Arab Emirates'),
    'af': _('Afghanistan'),
    'ag': _('Antigua & Barbuda'),
    'ai': _('Anguilla'),
    'al': _('Albania'),
    'am': _('Armenia'),
    'ao': _('Angola'),
    'aq': _('Antarctica'),
    'ar': _('Argentina'),
    'as': _('American Samoa'),
    'at': _('Austria'),
    'au': _('Australia'),
    'aw': _('Aruba'),
    'ax': _('Åland Islands'),
    'az': _('Azerbaijan'),
    'ba': _('Bosnia & Herzegovina'),
    'bb': _('Barbados'),
    'bd': _('Bangladesh'),
    'be': _('Belgium'),
    'bf': _('Burkina Faso'),
    'bg': _('Bulgaria'),
    'bh': _('Bahrain'),
    'bi': _('Burundi'),
    'bj': _('Benin'),
    'bl': _('Saint Barthelemy'),
    'bm': _('Bermuda'),
    'bn': _('Brunei Darussalam'),
    'bo': _('Bolivia'),
    'bq': _('Bonaire), Sint Eustatius and Saba'),
    'br': _('Brazil'),
    'bs': _('Bahamas'),
    'bt': _('Bhutan'),
    'bv': _('Bouvet Island'),
    'bw': _('Botswana'),
    'by': _('Belarus'),
    'bz': _('Belize'),
    'ca': _('Canada'),
    'cc': _('Cocos (Keeling) Islands'),
    'cd': _('Democratic Republic of Congo'),
    'cf': _('Central African Republic'),
    'cg': _('Congo'),
    'ch': _('Switzerland'),
    'ci': _('Ivory Coast'),
    'ck': _('Cook Islands'),
    'cl': _('Chile'),
    'cm': _('Cameroon'),
    'cn': _('China'),
    'co': _('Colombia'),
    'cr': _('Costa Rica'),
    'cu': _('Cuba'),
    'cv': _('Cabo Verde'),
    'cw': _('Curaçao'),
    'cx': _('Christmas Island'),
    'cy': _('Cyprus'),
    'cz': _('Czech Republic'),
    'de': _('Germany'),
    'dj': _('Djibouti'),
    'dk': _('Denmark'),
    'dm': _('Dominica'),
    'do': _('Dominican Republic'),
    'dz': _('Algeria'),
    'ec': _('Ecuador'),
    'ee': _('Estonia'),
    'eg': _('Egypt'),
    'eh': _('Western Sahara'),
    'er': _('Eritrea'),
    'es': _('Spain'),
    'et': _('Ethiopia'),
    'eu': _('Europe'),
    'fi': _('Finland'),
    'fj': _('Fiji'),
    'fk': _('Falkland Islands (Malvinas)'),
    'fm': _('Micronesia'),
    'fo': _('Faroe Islands'),
    'fr': _('France'),
    'ga': _('Gabon'),
    'gb': _('Great Britain'),
    'gd': _('Grenada'),
    'ge': _('Georgia'),
    'gf': _('French Guiana'),
    'gg': _('Guernsey'),
    'gh': _('Ghana'),
    'gi': _('Gibraltar'),
    'gl': _('Greenland'),
    'gm': _('Gambia'),
    'gn': _('Guinea'),
    'gp': _('Guadeloupe'),
    'gq': _('Equatorial Guinea'),
    'gr': _('Greece'),
    'gs': _('South Georgia & South Sandwich Islands'),
    'gt': _('Guatemala'),
    'gu': _('Guam'),
    'gw': _('Guinea-Bissau'),
    'gy': _('Guyana'),
    'hk': _('Hong Kong'),
    'hm': _('Heard & McDonald Islands'),
    'hn': _('Honduras'),
    'hr': _('Croatia'),
    'ht': _('Haiti'),
    'hu': _('Hungary'),
    'id': _('Indonesia'),
    'ie': _('Ireland'),
    'il': _('Israel'),
    'im': _('Isle of Man'),
    'in': _('India'),
    'io': _('British Indian Ocean Territory'),
    'iq': _('Iraq'),
    'ir': _('Iran'),
    'is': _('Iceland'),
    'it': _('Italy'),
    'je': _('Jersey'),
    'jm': _('Jamaica'),
    'jo': _('Jordan'),
    'jp': _('Japan'),
    'ke': _('Kenya'),
    'kg': _('Kyrgyzstan'),
    'kh': _('Cambodia'),
    'ki': _('Kiribati'),
    'km': _('Comoros'),
    'kn': _('Saint Kitts & Nevis'),
    'kp': _('North Korea'),
    'kr': _('South Korea'),
    'kw': _('Kuwait'),
    'ky': _('Cayman Islands'),
    'kz': _('Kazakhstan'),
    'la': _('Laos'),
    'lb': _('Lebanon'),
    'lc': _('Saint Lucia'),
    'li': _('Liechtenstein'),
    'lk': _('Sri Lanka'),
    'lr': _('Liberia'),
    'ls': _('Lesotho'),
    'lt': _('Lithuania'),
    'lu': _('Luxembourg'),
    'lv': _('Latvia'),
    'ly': _('Libya'),
    'ma': _('Morocco'),
    'mc': _('Monaco'),
    'md': _('Moldova'),
    'me': _('Montenegro'),
    'mf': _('Saint Martin'),
    'mg': _('Madagascar'),
    'mh': _('Marshall Islands'),
    'mk': _('North Macedonia'),
    'ml': _('Mali'),
    'mm': _('Myanmar'),
    'mn': _('Mongolia'),
    'mo': _('Macau'),
    'mp': _('Northern Mariana Islands'),
    'mq': _('Martinique'),
    'mr': _('Mauritania'),
    'ms': _('Montserrat'),
    'mt': _('Malta'),
    'mu': _('Mauritius'),
    'mv': _('Maldives'),
    'mw': _('Malawi'),
    'mx': _('Mexico'),
    'my': _('Malaysia'),
    'mz': _('Mozambique'),
    'na': _('Namibia'),
    'nc': _('New Caledonia'),
    'ne': _('Niger'),
    'nf': _('Norfolk Island'),
    'ng': _('Nigeria'),
    'ni': _('Nicaragua'),
    'nl': _('Netherlands'),
    'no': _('Norway'),
    'np': _('Nepal'),
    'nr': _('Nauru'),
    'nu': _('Niue'),
    'nz': _('New Zealand'),
    'om': _('Oman'),
    'pa': _('Panama'),
    'pe': _('Peru'),
    'pf': _('French Polynesia'),
    'pg': _('Papua New Guinea'),
    'ph': _('Philippines'),
    'pk': _('Pakistan'),
    'pl': _('Poland'),
    'pm': _('Saint Pierre & Miquelon'),
    'pn': _('Pitcairn'),
    'pr': _('Puerto Rico'),
    'ps': _('State of Palestine'),
    'pt': _('Portugal'),
    'pw': _('Palau'),
    'py': _('Paraguay'),
    'qa': _('Qatar'),
    're': _('Reunion'),
    'ro': _('Romania'),
    'rs': _('Serbia'),
    'ru': _('Russia'),
    'rw': _('Rwanda'),
    'sa': _('Saudi Arabia'),
    'sb': _('Solomon Islands'),
    'sc': _('Seychelles'),
    'sd': _('Sudan'),
    'se': _('Sweden'),
    'sg': _('Singapore'),
    'sh': _('Saint Helena'),
    'si': _('Slovenia'),
    'sj': _('Svalbard & Jan Mayen Islands'),
    'sk': _('Slovak Republic'),
    'sl': _('Sierra Leone'),
    'sm': _('San Marino'),
    'sn': _('Senegal'),
    'so': _('Somalia'),
    'sr': _('Suriname'),
    'ss': _('South Sudan'),
    'st': _('Sao Tome & Principe'),
    'sv': _('El Salvador'),
    'sx': _('Sint Maarten'),
    'sy': _('Syria'),
    'sz': _('Eswatini'),
    'tc': _('Turks & Caicos Islands'),
    'td': _('Chad'),
    'tf': _('French Southern Territories'),
    'tg': _('Togo'),
    'th': _('Thailand'),
    'tj': _('Tajikistan'),
    'tk': _('Tokelau'),
    'tl': _('Timor-Leste'),
    'tm': _('Turkmenistan'),
    'tn': _('Tunisia'),
    'to': _('Tonga'),
    'tr': _('Turkey'),
    'tt': _('Trinidad & Tobago'),
    'tv': _('Tuvalu'),
    'tw': _('Taiwan'),
    'tz': _('Tanzania'),
    'ua': _('Ukraine'),
    'ug': _('Uganda'),
    'um': _('U.S. Minor Outlying Islands'),
    'us': _('United States'),
    'uy': _('Uruguay'),
    'uz': _('Uzbekistan'),
    'va': _('Holy See (Vatican City State)'),
    'vc': _('Saint Vincent & The Grenadines'),
    've': _('Venezuela'),
    'vg': _('British Virgin Islands'),
    'vi': _('U.S. Virgin Islands'),
    'vn': _('Viet Nam'),
    'vu': _('Vanuatu'),
    'wf': _('Wallis & Futuna'),
    'ws': _('Samoa'),
    'xk': _('Kosovo'),
    'ye': _('Yemen'),
    'yt': _('Mayotte'),
    'za': _('South Africa'),
    'zm': _('Zambia'),
    'zw': _('Zimbabwe')
}



from gi.repository import Gio
from pynicotine.pluginsystem import BasePlugin


def enable(plugins):
    global PLUGIN
    PLUGIN = Plugin(plugins)
    PLUGIN.add_mpris_signal_receiver()


def disable(plugins):
    global PLUGIN
    PLUGIN.remove_mpris_signal_receiver()
    PLUGIN = None


class Plugin(BasePlugin):
    __name__ = "Now Playing Sender"
    settings = {
        'rooms': ['testroom'],
    }
    metasettings = {
        'rooms': {'description': 'Rooms to broadcast in', 'type': 'list string'},
    }

    def init(self):

        self.last_song_url = ""
        self.stop = False

        self.bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        self.signal_id = None

        self.dbus_mpris_service = 'org.mpris.MediaPlayer2.'
        self.dbus_mpris_player_service = 'org.mpris.MediaPlayer2.Player'
        self.dbus_mpris_path = '/org/mpris/MediaPlayer2'
        self.dbus_property = 'org.freedesktop.DBus.Properties'

    def add_mpris_signal_receiver(self):
        """ Receive updates related to MPRIS """

        self.signal_id = self.bus.signal_subscribe(None,
                                                   self.dbus_property,
                                                   "PropertiesChanged",
                                                   self.dbus_mpris_path,
                                                   None,
                                                   Gio.DBusSignalFlags.NONE,
                                                   self.song_change)

    def remove_mpris_signal_receiver(self):
        """ Stop receiving updates related to MPRIS """

        self.bus.signal_unsubscribe(self.signal_id)

    def get_current_mpris_player(self):
        """ Returns the MPRIS client currently selected in Now Playing """

        player = self.frame.np.config.sections["players"]["npothercommand"]

        if not player:
            dbus_proxy = Gio.DBusProxy.new_sync(self.bus,
                                                Gio.DBusProxyFlags.NONE,
                                                None,
                                                'org.freedesktop.DBus',
                                                '/org/freedesktop/DBus',
                                                'org.freedesktop.DBus',
                                                None)

            names = dbus_proxy.ListNames()

            for name in names:
                if name.startswith(self.dbus_mpris_service):
                    player = name[len(self.dbus_mpris_service):]
                    break

        return player

    def get_current_mpris_song_url(self, player):
        """ Returns the current song url for the selected MPRIS client """

        dbus_proxy = Gio.DBusProxy.new_sync(self.bus,
                                            Gio.DBusProxyFlags.NONE,
                                            None,
                                            self.dbus_mpris_service + player,
                                            self.dbus_mpris_path,
                                            self.dbus_property,
                                            None)

        metadata = dbus_proxy.Get('(ss)', self.dbus_mpris_player_service, 'Metadata')
        song_url = metadata.get("xesam:url")

        return song_url

    def send_now_playing(self):
        """ Broadcast Now Playing in selected rooms """

        for room in self.settings['rooms']:
            playing = self.frame.np.now_playing.get_np()

            if playing:
                self.saypublic(room, playing)

    def song_change(self, connection, sender_name, object_path, interface_name, signal_name, parameters):

        if self.frame.np.config.sections["players"]["npplayer"] != "mpris":
            # MPRIS is not active, exit
            return

        # Get the changed song url received from the the signal
        try:
            changed_song_url = parameters[1].get("Metadata").get("xesam:url")

        except AttributeError:
            return

        if not changed_song_url:
            # Song url empty, the player most likely stopped playing
            self.last_song_url = ""
            return

        if changed_song_url == self.last_song_url:
            # A new song didn't start playing, exit
            return

        try:
            player = self.get_current_mpris_player()
            selected_client_song_url = self.get_current_mpris_song_url(player)

        except Exception:
            # Selected player is invalid
            return

        if selected_client_song_url != changed_song_url:
            # Song change was from another MPRIS client than the selected one, exit
            return

        # Keep track of which song is playing
        self.last_song_url = changed_song_url

        self.send_now_playing()



# COPYRIGHT (C) 2020-2021 Nicotine+ Team
# COPYRIGHT (C) 2020 Mathias <mail@mathias.is>
# COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
# COPYRIGHT (C) 2016 Mutnick <muhing@yahoo.com>
# COPYRIGHT (C) 2013 eL_vErDe <gandalf@le-vert.net>
# COPYRIGHT (C) 2008-2012 Quinox <quinox@users.sf.net>
# COPYRIGHT (C) 2009 Hedonist <ak@sensi.org>
# COPYRIGHT (C) 2006-2009 Daelstorm <daelstorm@gmail.com>
# COPYRIGHT (C) 2003-2004 Hyriand <hyriand@thegraveyard.org>
# COPYRIGHT (C) 2001-2003 Alexander Kanavin
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This is the actual client code. Actual GUI classes are in the separate modules
"""

import os
import queue
import threading
import time
import _thread

from pynicotine import slskmessages
from pynicotine import slskproto
from pynicotine import transfers
from pynicotine.config import Config
from pynicotine.geoip.ip2location import IP2Location
from pynicotine.logfacility import log
from pynicotine.nowplaying import NowPlaying
from pynicotine.shares import Shares
from pynicotine.slskmessages import new_id
from pynicotine.transfers import Statistics
from pynicotine.utils import clean_file
from pynicotine.utils import unescape


class PeerConnection:
    """
    Holds information about a peer connection. Not every field may be set
    to something. addr is (ip, port) address, conn is a socket object, msgs is
    a list of outgoing pending messages, token is a reverse-handshake
    number (protocol feature), init is a PeerInit protocol message. (read
    slskmessages docstrings for explanation of these)
    """

    __slots__ = "addr", "username", "conn", "msgs", "token", "init", "type", "conntimer", "tryaddr"

    def __init__(self, addr=None, username=None, conn=None, msgs=None, token=None, init=None, conntimer=None, tryaddr=None):
        self.addr = addr
        self.username = username
        self.conn = conn
        self.msgs = msgs
        self.token = token
        self.init = init
        self.type = init.type
        self.conntimer = conntimer
        self.tryaddr = tryaddr


class Timeout:

    __slots__ = "callback"

    def __init__(self, callback):
        self.callback = callback

    def timeout(self):
        try:
            self.callback([self])
        except Exception as e:
            log.add_warning(_("Exception in callback %s: %s"), (self.callback, e))


class ConnectToPeerTimeout(Timeout):

    __slots__ = "conn"

    def __init__(self, conn, callback):
        self.conn = conn
        self.callback = callback


class NetworkEventProcessor:
    """ This class contains handlers for various messages from the networking thread """

    def __init__(self, ui_callback, network_callback, setstatus, bindip, port, data_dir, config):

        self.ui_callback = ui_callback
        self.network_callback = network_callback
        self.set_status = setstatus
        self.manualdisconnect = False

        try:
            self.config = Config(config, data_dir)

        except Exception:
            import shutil
            corruptfile = ".".join([config, clean_file(time.strftime("%Y-%m-%d_%H_%M_%S")), "corrupt"])
            shutil.move(config, corruptfile)
            short = _("Your config file is corrupt")
            long = _("We're sorry, but it seems your configuration file is corrupt. Please reconfigure Nicotine+.\n\nWe renamed your old configuration file to\n%(corrupt)s\nIf you open this file with a text editor you might be able to rescue some of your settings.") % {'corrupt': corruptfile}
            self.config = Config(config, data_dir)
            self.network_callback([slskmessages.PopupMessage(short, long)])

        self.bindip = bindip
        self.port = port

        log.set_log_levels(self.config.sections["logging"]["debugmodes"])

        self.peerconns = []
        self.watchedusers = set()
        self.ipblock_requested = {}
        self.ipignore_requested = {}
        self.ip_requested = set()
        self.private_message_queue = {}
        self.users = {}

        self.queue = queue.Queue(0)
        self.statistics = Statistics(self.config, self.ui_callback)
        self.shares = Shares(self, self.config, self.queue, self.ui_callback)
        self.pluginhandler = None  # Initialized when the GUI is ready
        self.now_playing = NowPlaying(self.config)

        script_dir = os.path.dirname(__file__)
        file_path = os.path.join(script_dir, "geoip/ipcountrydb.bin")
        self.geoip = IP2Location(file_path, "SHARED_MEMORY")

        # Give the logger information about log folder
        self.update_debug_log_options()

        self.protothread = slskproto.SlskProtoThread(self.network_callback, self.queue, self.bindip, self.port, self.config, self)

        # UPnP
        self.upnp_interval = self.config.sections["server"]["upnp_interval"]
        self.upnp_timer = None
        _thread.start_new_thread(self.add_upnp_portmapping, ())

        self.active_server_conn = None
        self.waitport = None
        self.chatrooms = None
        self.privatechat = None
        self.globallist = None
        self.userinfo = None
        self.userbrowse = None
        self.search = None
        self.transfers = None
        self.userlist = None
        self.ipaddress = None
        self.privileges_left = None
        self.servertimer = None
        self.server_timeout_value = -1

        self.has_parent = False

        self.requested_info = {}
        self.requested_folders = {}
        self.speed = 0

        # Callback handlers for messages
        self.events = {
            slskmessages.ConnectToServer: self.connect_to_server,
            slskmessages.ConnectError: self.connect_error,
            slskmessages.IncPort: self.inc_port,
            slskmessages.ServerConn: self.server_conn,
            slskmessages.ConnClose: self.conn_close,
            slskmessages.Login: self.login,
            slskmessages.ChangePassword: self.change_password,
            slskmessages.MessageUser: self.message_user,
            slskmessages.PMessageUser: self.p_message_user,
            slskmessages.ExactFileSearch: self.dummy_message,
            slskmessages.RoomAdded: self.dummy_message,
            slskmessages.RoomRemoved: self.dummy_message,
            slskmessages.UserJoinedRoom: self.user_joined_room,
            slskmessages.SayChatroom: self.say_chat_room,
            slskmessages.JoinRoom: self.join_room,
            slskmessages.UserLeftRoom: self.user_left_room,
            slskmessages.CantCreateRoom: self.dummy_message,
            slskmessages.QueuedDownloads: self.dummy_message,
            slskmessages.GetPeerAddress: self.get_peer_address,
            slskmessages.OutConn: self.out_conn,
            slskmessages.UserInfoReply: self.user_info_reply,
            slskmessages.UserInfoRequest: self.user_info_request,
            slskmessages.PierceFireWall: self.pierce_fire_wall,
            slskmessages.CantConnectToPeer: self.cant_connect_to_peer,
            slskmessages.PeerTransfer: self.peer_transfer,
            slskmessages.SharedFileList: self.shared_file_list,
            slskmessages.GetSharedFileList: self.get_shared_file_list,
            slskmessages.FileSearchRequest: self.file_search_request,
            slskmessages.FileSearchResult: self.file_search_result,
            slskmessages.ConnectToPeer: self.connect_to_peer_request,
            slskmessages.GetUserStatus: self.get_user_status,
            slskmessages.GetUserStats: self.get_user_stats,
            slskmessages.Relogged: self.relogged,
            slskmessages.PeerInit: self.peer_init,
            slskmessages.CheckDownloadQueue: self.check_download_queue,
            slskmessages.DownloadFile: self.file_download,
            slskmessages.UploadFile: self.file_upload,
            slskmessages.FileRequest: self.file_request,
            slskmessages.TransferRequest: self.transfer_request,
            slskmessages.TransferResponse: self.transfer_response,
            slskmessages.QueueUpload: self.queue_upload,
            slskmessages.QueueFailed: self.queue_failed,
            slskmessages.UploadFailed: self.upload_failed,
            slskmessages.PlaceInQueue: self.place_in_queue,
            slskmessages.FileError: self.file_error,
            slskmessages.FolderContentsResponse: self.folder_contents_response,
            slskmessages.FolderContentsRequest: self.folder_contents_request,
            slskmessages.RoomList: self.room_list,
            slskmessages.LeaveRoom: self.leave_room,
            slskmessages.GlobalUserList: self.global_user_list,
            slskmessages.AddUser: self.add_user,
            slskmessages.PrivilegedUsers: self.privileged_users,
            slskmessages.AddToPrivileged: self.add_to_privileged,
            slskmessages.CheckPrivileges: self.check_privileges,
            slskmessages.ServerPing: self.dummy_message,
            slskmessages.ParentMinSpeed: self.dummy_message,
            slskmessages.ParentSpeedRatio: self.dummy_message,
            slskmessages.ParentInactivityTimeout: self.dummy_message,
            slskmessages.SearchInactivityTimeout: self.dummy_message,
            slskmessages.MinParentsInCache: self.dummy_message,
            slskmessages.WishlistInterval: self.wishlist_interval,
            slskmessages.DistribAliveInterval: self.dummy_message,
            slskmessages.ChildDepth: self.child_depth,
            slskmessages.BranchLevel: self.branch_level,
            slskmessages.BranchRoot: self.branch_root,
            slskmessages.DistribChildDepth: self.distrib_child_depth,
            slskmessages.DistribBranchLevel: self.distrib_branch_level,
            slskmessages.DistribBranchRoot: self.distrib_branch_root,
            slskmessages.AdminMessage: self.admin_message,
            slskmessages.TunneledMessage: self.tunneled_message,
            slskmessages.IncConn: self.inc_conn,
            slskmessages.PlaceholdUpload: self.dummy_message,
            slskmessages.PlaceInQueueRequest: self.place_in_queue_request,
            slskmessages.UploadQueueNotification: self.upload_queue_notification,
            slskmessages.SearchRequest: self.search_request,
            slskmessages.FileSearch: self.search_request,
            slskmessages.RoomSearch: self.room_search_request,
            slskmessages.UserSearch: self.search_request,
            slskmessages.RelatedSearch: self.dummy_message,
            slskmessages.PossibleParents: self.possible_parents,
            slskmessages.DistribAlive: self.dummy_message,
            slskmessages.DistribSearch: self.distrib_search,
            slskmessages.DistribServerSearch: self.distrib_search,
            ConnectToPeerTimeout: self.connect_to_peer_timeout,
            transfers.TransferTimeout: self.transfer_timeout,
            str: self.notify,
            slskmessages.PopupMessage: self.popup_message,
            slskmessages.SetCurrentConnectionCount: self.set_current_connection_count,
            slskmessages.GlobalRecommendations: self.global_recommendations,
            slskmessages.Recommendations: self.recommendations,
            slskmessages.ItemRecommendations: self.item_recommendations,
            slskmessages.SimilarUsers: self.similar_users,
            slskmessages.ItemSimilarUsers: self.similar_users,
            slskmessages.UserInterests: self.user_interests,
            slskmessages.RoomTickerState: self.room_ticker_state,
            slskmessages.RoomTickerAdd: self.room_ticker_add,
            slskmessages.RoomTickerRemove: self.room_ticker_remove,
            slskmessages.UserPrivileged: self.user_privileged,
            slskmessages.AckNotifyPrivileges: self.ack_notify_privileges,
            slskmessages.NotifyPrivileges: self.notify_privileges,
            slskmessages.PrivateRoomUsers: self.private_room_users,
            slskmessages.PrivateRoomOwned: self.private_room_owned,
            slskmessages.PrivateRoomAddUser: self.private_room_add_user,
            slskmessages.PrivateRoomRemoveUser: self.private_room_remove_user,
            slskmessages.PrivateRoomAdded: self.private_room_added,
            slskmessages.PrivateRoomRemoved: self.private_room_removed,
            slskmessages.PrivateRoomDisown: self.private_room_disown,
            slskmessages.PrivateRoomToggle: self.private_room_toggle,
            slskmessages.PrivateRoomSomething: self.dummy_message,
            slskmessages.PrivateRoomOperatorAdded: self.private_room_operator_added,
            slskmessages.PrivateRoomOperatorRemoved: self.private_room_operator_removed,
            slskmessages.PrivateRoomAddOperator: self.private_room_add_operator,
            slskmessages.PrivateRoomRemoveOperator: self.private_room_remove_operator,
            slskmessages.PublicRoomMessage: self.public_room_message,
            slskmessages.UnknownPeerMessage: self.ignore
        }

    def peer_init(self, msg):

        """ Peer wants to connect to us, remember them """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        user = msg.user
        addr = msg.conn.addr
        conn = msg.conn.conn
        msg_type = msg.type
        found_conn = False

        if user != self.config.sections["server"]["login"]:  # We need two connections in our name if we're downloading from ourselves
            for i in self.peerconns:
                if i.username == user and i.type != 'F' and i.type == msg_type:
                    i.addr = addr
                    i.conn = conn
                    i.token = None
                    i.init = msg

                    if i.conntimer is not None:
                        i.conntimer.cancel()

                    found_conn = True
                    self.process_conn_messages(i, conn)
                    break

        if not found_conn:
            """ No previous connection exists for user """

            self.peerconns.append(
                PeerConnection(
                    addr=addr,
                    username=user,
                    conn=conn,
                    init=msg,
                    msgs=[]
                )
            )

        log.add_conn("Received incoming connection of type %(type)s from user %(user)s" % {'type': msg_type, 'user': user})

    def send_message_to_peer(self, user, message, address=None):

        """ Sends message to a peer. Used primarily when we know the username of a peer,
        but don't have an active connection. """

        conn = None

        if message.__class__ is not slskmessages.FileRequest:
            """ Check if there's already a connection object for the specified username """

            for i in self.peerconns:
                if i.username == user and i.type == 'P':
                    conn = i
                    log.add_conn("Found existing connection of type %(type)s for user %(user)s, using it." % {
                        'type': i.type,
                        'user': user
                    })
                    break

        if conn is not None and conn.conn is not None:
            """ We have initiated a connection previously, and it's ready """

            message.conn = conn.conn
            self.queue.put(message)

        elif conn is not None:
            """ Connection exists but is not ready yet, add new messages to it """

            conn.msgs.append(message)

        else:
            """ This is a new peer, initiate a connection """

            self.initiate_connection_to_peer(user, message, address)

        log.add_conn("Sending message of type %(type)s to user %(user)s" % {
            'type': message.__class__,
            'user': user
        })

    def initiate_connection_to_peer(self, user, message, address=None):

        """ Prepare to initiate a connection with a peer """

        if message.__class__ is slskmessages.FileRequest:
            message_type = 'F'

        elif message.__class__ is slskmessages.DistribConn:
            message_type = 'D'

        else:
            message_type = 'P'

        init = slskmessages.PeerInit(None, self.config.sections["server"]["login"], message_type, 0)
        addr = None

        if user in self.users:
            addr = self.users[user].addr

        elif address is not None:
            self.users[user] = UserAddr(status=-1, addr=address)
            addr = address

        if addr is None:
            self.queue.put(slskmessages.GetPeerAddress(user))

            if message.__class__ is slskmessages.TransferRequest and self.transfers is not None:
                self.transfers.getting_address(message.req, message.direction)

            log.add_conn("Requesting address for user %(user)s", {
                'user': user
            })

        else:
            self.connect_to_peer_direct(user, addr, message_type)

        self.peerconns.append(
            PeerConnection(
                addr=addr,
                username=user,
                msgs=[message],
                init=init
            )
        )

    def connect_to_peer_direct(self, user, addr, message_type, init=None):

        """ Initiate a connection with a peer directly """

        self.queue.put(slskmessages.OutConn(None, addr, init))

        log.add_conn("Initialising direct connection of type %(type)s to user %(user)s", {
            'type': message_type,
            'user': user
        })

    def connect_to_peer_indirect(self, conn, error):

        """ Send a message to the server to ask the peer to connect to us instead (indirect connection) """

        conn.token = new_id()
        self.queue.put(slskmessages.ConnectToPeer(conn.token, conn.username, conn.type))

        for j in conn.msgs:
            if j.__class__ is slskmessages.TransferRequest and self.transfers is not None:
                self.transfers.got_connect_error(j.req, j.direction)

        conntimeout = ConnectToPeerTimeout(conn, self.network_callback)
        timer = threading.Timer(20.0, conntimeout.timeout)
        timer.setName("ConnectionTimer")
        timer.setDaemon(True)
        timer.start()

        if conn.conntimer is not None:
            conn.conntimer.cancel()

        conn.conntimer = timer

        log.add_conn(
            "Direct connection of type %(type)s to user %(user)s failed, attempting indirect connection. Error: %(error)s", {
                "type": conn.type,
                "user": conn.username,
                "error": error
            }
        )

    def connect_to_peer_request(self, msg):

        """ Peer sent us an indirect connection request via the server, attempt to
        connect to them """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        user = msg.user
        addr = (msg.ip, msg.port)
        token = msg.token
        msg_type = msg.type
        found_conn = False
        should_connect = True

        init = slskmessages.PeerInit(None, user, msg_type, 0)

        if user != self.config.sections["server"]["login"]:
            for i in self.peerconns:
                if i.username == user and i.type != 'F' and i.type == msg_type:
                    """ Only update existing connection if it hasn't been established yet,
                    otherwise ignore indirect connection request. """

                    found_conn = True

                    if i.conn is None:
                        i.addr = addr
                        i.token = token
                        i.init = init
                        break

                    if i.conntimer is not None:
                        i.conntimer.cancel()

                    should_connect = False
                    break

        if should_connect:
            self.connect_to_peer_direct(user, addr, msg_type, init)

        if not found_conn:
            """ No previous connection exists for user """

            self.peerconns.append(
                PeerConnection(
                    addr=addr,
                    username=user,
                    msgs=[],
                    token=token,
                    init=init
                )
            )

    def get_peer_address(self, msg):

        """ Server responds with the IP address and port of the user we requested """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        user = msg.user

        for i in self.peerconns:
            if i.username == user and i.addr is None:
                if msg.port != 0 or i.tryaddr == 10:

                    """ We now have the IP address for a user we previously didn't know,
                    attempt a direct connection to the peer/user """

                    if i.tryaddr == 10:
                        log.add_conn(
                            "Server reported port 0 for the 10th time for user %(user)s, giving up", {
                                'user': user
                            }
                        )

                    elif i.tryaddr is not None:
                        log.add_conn(
                            "Server reported non-zero port for user %(user)s after %(tries)i retries", {
                                'user': user,
                                'tries': i.tryaddr
                            }
                        )

                    i.addr = (msg.ip, msg.port)
                    i.tryaddr = None

                    self.connect_to_peer_direct(user, i.addr, i.type)

                    for j in i.msgs:
                        if j.__class__ is slskmessages.TransferRequest and self.transfers is not None:
                            self.transfers.got_address(j.req, j.direction)
                else:

                    """ Server responded with an incorrect port, request peer address again """

                    if i.tryaddr is None:
                        i.tryaddr = 1
                        log.add_conn(
                            "Server reported port 0 for user %(user)s, retrying", {
                                'user': user
                            }
                        )
                    else:
                        i.tryaddr += 1

                    self.queue.put(slskmessages.GetPeerAddress(user))
                    return

        if user in self.users:
            self.users[user].addr = (msg.ip, msg.port)
        else:
            self.users[user] = UserAddr(addr=(msg.ip, msg.port))

        if user in self.ipblock_requested:

            if self.ipblock_requested[user]:
                self.ui_callback.on_un_block_user(user)
            else:
                self.ui_callback.on_block_user(user)

            del self.ipblock_requested[user]
            return

        if user in self.ipignore_requested:

            if self.ipignore_requested[user]:
                self.ui_callback.on_un_ignore_user(user)
            else:
                self.ui_callback.on_ignore_user(user)

            del self.ipignore_requested[user]
            return

        ip_record = self.geoip.get_all(msg.ip)
        cc = ip_record.country_short

        if cc == "-":
            cc = ""

        self.ui_callback.has_user_flag(user, "flag_" + cc)

        # From this point on all paths should call
        # self.pluginhandler.user_resolve_notification precisely once
        if user in self.private_message_queue:
            self.private_message_queue_process(user)

        if user not in self.ip_requested:
            self.pluginhandler.user_resolve_notification(user, msg.ip, msg.port)
            return

        self.ip_requested.remove(user)
        self.pluginhandler.user_resolve_notification(user, msg.ip, msg.port, cc)

        if cc != "":
            country = " (%(cc)s / %(country)s)" % {'cc': cc, 'country': ip_record.country_long}
        else:
            country = ""

        log.add(_("IP address of %(user)s is %(ip)s, port %(port)i%(country)s"), {
            'user': user,
            'ip': msg.ip,
            'port': msg.port,
            'country': country
        })

    def process_conn_messages(self, peerconn, conn):

        """ We may have messages from our peer, represent these in our UI """

        for j in peerconn.msgs:

            if j.__class__ is slskmessages.UserInfoRequest and self.userinfo is not None:
                self.userinfo.show_user(peerconn.username, conn=conn)

            if j.__class__ is slskmessages.GetSharedFileList and self.userbrowse is not None:
                self.userbrowse.show_user(peerconn.username, conn=conn)

            if j.__class__ is slskmessages.FileRequest and self.transfers is not None:
                self.transfers.got_file_connect(j.req, conn)

            if j.__class__ is slskmessages.TransferRequest and self.transfers is not None:
                self.transfers.got_connect(j.req, conn, j.direction)

            j.conn = conn
            self.queue.put(j)

        peerconn.msgs = []

    def inc_conn(self, msg):
        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

    def out_conn(self, msg):

        """ Direct connection to peer was successful, send pierce firewall message
        and process any messages associated with the connection """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        addr = msg.addr

        for i in self.peerconns:
            if i.addr == addr and i.conn is None:
                conn = msg.conn

                if i.token is None:
                    i.init.conn = conn
                    self.queue.put(i.init)

                else:
                    self.queue.put(slskmessages.PierceFireWall(conn, i.token))

                i.conn = conn

                log.add_conn("Connection established with user %(user)s, connection contains messages: %(messages)s" % {
                    'user': i.username,
                    'messages': i.msgs
                })

                # Update UI with contents from messages
                self.process_conn_messages(i, conn)

                break

    def pierce_fire_wall(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        # Sometimes a token is seemingly not sent by the peer, check if the
        # IP address matches in that case

        for i in self.peerconns:
            if i.token is None or i.conn is not None:
                continue

            if msg.token is None and i.addr == msg.conn.addr or \
                    i.token == msg.token:
                conn = msg.conn.conn

                if i.conntimer is not None:
                    i.conntimer.cancel()

                i.init.conn = conn
                self.queue.put(i.init)
                i.conn = conn

                log.add_conn("Received PierceFirewall message from user %(user)s, connection contains messages: %(messages)s" % {
                    'user': i.username,
                    'messages': i.msgs
                })

                # Update UI with contents from messages
                self.process_conn_messages(i, conn)

                break

    def close_peer_connection(self, conn, username):

        """ Forcibly close a peer connection. Only used after receiving a search result,
        as we need to get rid of connections before they pile up """

        if conn is None:
            return

        if not self.protothread.socket_still_active(conn.conn):

            for i in reversed(self.peerconns):  # File connections are added after peer connections
                if i.username == username and i.type == 'F':
                    # File transfer in progress, don't kill peer connection yet
                    return

                if i.conn == conn.conn:
                    self.peerconns.remove(i)
                    break

            self.queue.put(slskmessages.ConnClose(conn.conn, callback=False))

    def show_connection_error_message(self, conn):

        """ Request UI to show error messages related to connectivity """

        for i in conn.msgs:
            if i.__class__ in [slskmessages.TransferRequest, slskmessages.FileRequest] and self.transfers is not None:
                self.transfers.got_cant_connect(i.req)

            if i.__class__ is slskmessages.GetSharedFileList and self.userbrowse is not None:
                self.userbrowse.show_connection_error(conn.username)

            if i.__class__ is slskmessages.UserInfoRequest and self.userinfo is not None:
                self.userinfo.show_connection_error(conn.username)

    def cant_connect_to_peer(self, msg):

        """ Server informs us that an indirect connection with a peer has failed.
        Game over. """
        """ DEPRECATED. While we can still receive CantConnectToPeer messages by
        other clients, Nicotine+ does not send such messages anymore. """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        token = msg.token

        for i in self.peerconns:
            if i.token == token:

                if i.conntimer is not None:
                    i.conntimer.cancel()

                self.peerconns.remove(i)

                self.show_connection_error_message(i)
                log.add_conn("Can't connect to user %s neither directly nor indirectly, giving up", i.username)
                break

    def connect_to_peer_timeout(self, msg):

        """ Peer was not able to repond to our indirect connection request """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        conn = msg.conn

        try:
            self.peerconns.remove(conn)
        except ValueError:
            pass

        self.show_connection_error_message(conn)
        log.add_conn(
            "Indirect connect request of type %(type)s to user %(user)s expired, giving up", {
                'type': conn.type,
                'user': conn.username
            }
        )

    def closed_connection(self, conn, addr, error=None):

        if conn == self.active_server_conn:

            self.set_status(
                _("Disconnected from server %(host)s:%(port)s"), {
                    'host': addr[0],
                    'port': addr[1]
                }
            )
            userchoice = self.manualdisconnect

            if not self.manualdisconnect:
                self.set_server_timer()
            else:
                self.manualdisconnect = False

            self.active_server_conn = None

            # Clean up connections
            self.peerconns.clear()

            self.watchedusers.clear()
            self.shares.set_connected(False)

            if self.transfers is not None:
                self.transfers.disconnect()

            self.privatechat = self.chatrooms = self.userinfo = self.userbrowse = self.search = self.transfers = self.userlist = None
            self.ui_callback.conn_close(conn, addr)
            self.pluginhandler.server_disconnect_notification(userchoice)

        else:

            """ A peer connection has died, remove it """

            for i in self.peerconns:
                if i.conn == conn:
                    log.add_conn("Connection closed by peer: %(peer)s. Error: %(error)s", {'peer': self.contents(i), 'error': error})

                    if i.conntimer is not None:
                        i.conntimer.cancel()

                    if self.transfers is not None:
                        self.transfers.conn_close(conn, addr, i.username, error)

                    if i.type == 'D':
                        self.parent_conn_closed()

                    self.peerconns.remove(i)
                    return

    def conn_close(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        self.closed_connection(msg.conn, msg.addr)

    def connect_error(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if msg.connobj.__class__ is slskmessages.ServerConn:

            self.set_status(
                _("Can't connect to server %(host)s:%(port)s: %(error)s"), {
                    'host': msg.connobj.addr[0],
                    'port': msg.connobj.addr[1],
                    'error': msg.err
                }
            )

            self.set_server_timer()

            if self.active_server_conn is not None:
                self.active_server_conn = None

            self.ui_callback.connect_error(msg)

        elif msg.connobj.__class__ is slskmessages.OutConn:

            addr = msg.connobj.addr

            for i in self.peerconns:
                if i.addr == addr and i.conn is None:
                    if i.token is None:

                        """ We can't correct to peer directly, request indirect connection """

                        self.connect_to_peer_indirect(i, msg.err)

                    elif i.conntimer is None:

                        """ Peer sent us an indirect connection request, and we weren't able to
                        connect to them. """

                        for j in i.msgs:
                            if j.__class__ in (slskmessages.TransferRequest, slskmessages.FileRequest) and self.transfers is not None:
                                self.transfers.got_cant_connect(j.req)

                        log.add_conn("Can't connect to user %(user)s indirectly. Error: %(error)s", {
                            'user': i.username,
                            'error': msg.err
                        })

                        self.peerconns.remove(i)

                    break

        else:
            self.closed_connection(msg.connobj.conn, msg.connobj.addr, msg.err)

    def start_upnp_timer(self):
        """ Port mapping entries last 24 hours, we need to regularly renew them """
        """ The default interval is 4 hours """

        if self.upnp_interval < 1:
            return

        upnp_interval_seconds = self.upnp_interval * 60 * 60

        self.upnp_timer = threading.Timer(upnp_interval_seconds, self.add_upnp_portmapping)
        self.upnp_timer.setName("UPnPTimer")
        self.upnp_timer.setDaemon(True)
        self.upnp_timer.start()

    def add_upnp_portmapping(self):

        # Repeat
        self.start_upnp_timer()

        # Test if we want to do a port mapping
        if self.config.sections["server"]["upnp"]:

            # Initialise a UPnPPortMapping object
            from pynicotine.upnp.portmapper import UPnPPortMapping
            upnp = UPnPPortMapping()

            # Do the port mapping
            upnp.add_port_mapping(self)

    def set_server_timer(self):

        if self.server_timeout_value == -1:
            self.server_timeout_value = 15
        elif 0 < self.server_timeout_value < 600:
            self.server_timeout_value = self.server_timeout_value * 2

        self.servertimer = threading.Timer(self.server_timeout_value, self.server_timeout)
        self.servertimer.setName("ServerTimer")
        self.servertimer.setDaemon(True)
        self.servertimer.start()

        self.set_status(_("The server seems to be down or not responding, retrying in %i seconds"), (self.server_timeout_value))

    def server_timeout(self):
        if not self.config.need_config():
            self.network_callback([slskmessages.ConnectToServer()])

    def stop_timers(self):

        for i in self.peerconns:
            if i.conntimer is not None:
                i.conntimer.cancel()

        if self.servertimer is not None:
            self.servertimer.cancel()

        if self.transfers is not None:
            self.transfers.abort_transfers()

    def connect_to_server(self, msg):
        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))
        self.ui_callback.on_connect()

    # notify user of error when recieving or sending a message
    # @param self NetworkEventProcessor (Class)
    # @param string a string containing an error message
    def notify(self, string):
        log.add_msg_contents("%s", string)

    def contents(self, obj):
        """ Returns variables for object, for debug output """
        try:
            return {s: getattr(obj, s) for s in obj.__slots__ if hasattr(obj, s)}
        except AttributeError:
            return vars(obj)

    def popup_message(self, msg):
        self.set_status(_(msg.title))
        self.ui_callback.popup_message(msg)

    def set_current_connection_count(self, msg):
        self.ui_callback.set_socket_status(msg.msg)

    def inc_port(self, msg):
        self.waitport = msg.port
        self.set_status(_("Listening on port %i"), msg.port)

    def server_conn(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        self.set_status(
            _("Connected to server %(host)s:%(port)s, logging in..."), {
                'host': msg.addr[0],
                'port': msg.addr[1]
            }
        )

        self.active_server_conn = msg.conn
        self.server_timeout_value = -1
        self.users = {}
        self.queue.put(
            slskmessages.Login(
                self.config.sections["server"]["login"],
                self.config.sections["server"]["passw"],

                # Soulseek client version; 155, 156, 157
                # SoulseekQt seems to be using 157
                # We use a custom version number for Nicotine+
                160,

                # Soulseek client minor version
                # 17 stands for 157 ns 13c, 19 for 157 ns 13e
                # SoulseekQt seems to go higher than this
                # We use a custom minor version for Nicotine+
                1,
            )
        )
        if self.waitport is not None:
            self.queue.put(slskmessages.SetWaitPort(self.waitport))

    def login(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if msg.success:

            self.transfers = transfers.Transfers(self.peerconns, self.queue, self, self.users,
                                                 self.network_callback, self.ui_callback.notifications, self.pluginhandler)
            self.shares.set_connected(True)

            if msg.ip is not None:
                self.ipaddress = msg.ip

            self.privatechat, self.chatrooms, self.userinfo, self.userbrowse, self.search, downloads, uploads, self.userlist, self.interests = self.ui_callback.init_interface(msg)

            self.transfers.set_transfer_views(downloads, uploads)
            self.shares.send_num_shared_folders_files()
            self.queue.put(slskmessages.SetStatus((not self.ui_callback.away) + 1))

            for thing in self.config.sections["interests"]["likes"]:
                if thing and isinstance(thing, str):
                    self.queue.put(slskmessages.AddThingILike(thing))

            for thing in self.config.sections["interests"]["dislikes"]:
                if thing and isinstance(thing, str):
                    self.queue.put(slskmessages.AddThingIHate(thing))

            self.queue.put(slskmessages.HaveNoParent(1))

            """ TODO: Nicotine+ can currently receive search requests from a parent connection, but
            redirecting results to children is not implemented yet. Tell the server we don't accept
            children for now. """
            self.queue.put(slskmessages.AcceptChildren(0))

            self.queue.put(slskmessages.NotifyPrivileges(1, self.config.sections["server"]["login"]))
            self.privatechat.login()
            self.queue.put(slskmessages.CheckPrivileges())

            """ Request a complete room list. A limited room list not including blacklisted rooms and
            rooms with few users is automatically sent when logging in, but subsequent room list
            requests contain all rooms. """
            self.queue.put(slskmessages.RoomList())

            self.queue.put(slskmessages.PrivateRoomToggle(self.config.sections["server"]["private_chatrooms"]))
        else:
            self.manualdisconnect = True
            self.set_status(_("Can not log in, reason: %s"), (msg.reason))

    def change_password(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        password = msg.password
        self.config.sections["server"]["passw"] = password
        self.config.write_configuration()
        self.network_callback([slskmessages.PopupMessage(_("Your password has been changed"), "Password is %s" % password)])

    def notify_privileges(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if msg.token is not None:
            pass

    def user_privileged(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            if msg.privileged is True:
                self.transfers.add_to_privileged(msg.user)

    def ack_notify_privileges(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if msg.token is not None:
            # Until I know the syntax, sending this message is probably a bad idea
            self.queue.put(slskmessages.AckNotifyPrivileges(msg.token))

    def p_message_user(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        conn = msg.conn.conn
        user = None

        # Get peer's username
        for i in self.peerconns:
            if i.conn is conn:
                user = i.username
                break

        if user is None:
            # No peer connection
            return

        if user != msg.user:
            text = _("(Warning: %(realuser)s is attempting to spoof %(fakeuser)s) ") % {"realuser": user, "fakeuser": msg.user} + msg.msg
            msg.user = user
        else:
            text = msg.msg

        if self.privatechat is not None:
            self.privatechat.show_message(msg, text)

    def message_user(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.privatechat is not None:
            event = self.pluginhandler.incoming_private_chat_event(msg.user, msg.msg)

            if event is not None:
                (u, msg.msg) = event
                self.privatechat.show_message(msg, msg.msg, msg.newmessage)

                self.pluginhandler.incoming_private_chat_notification(msg.user, msg.msg)

            self.queue.put(slskmessages.MessageAcked(msg.msgid))

    def user_joined_room(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.user_joined_room(msg)

    def public_room_message(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.public_room_message(msg, msg.msg)
            self.pluginhandler.public_room_message_notification(msg.room, msg.user, msg.msg)

    def join_room(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.join_room(msg)

    def private_room_users(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_users(msg)

    def private_room_owned(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_owned(msg)

    def private_room_add_user(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_add_user(msg)

    def private_room_remove_user(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_remove_user(msg)

    def private_room_operator_added(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_operator_added(msg)

    def private_room_operator_removed(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_operator_removed(msg)

    def private_room_add_operator(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_add_operator(msg)

    def private_room_remove_operator(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_remove_operator(msg)

    def private_room_added(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_added(msg)

    def private_room_removed(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_removed(msg)

    def private_room_disown(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.private_room_disown(msg)

    def private_room_toggle(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.toggle_private_rooms(msg.enabled)

    def leave_room(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.leave_room(msg)

    def private_message_queue_add(self, msg, text):

        user = msg.user

        if user not in self.private_message_queue:
            self.private_message_queue[user] = [[msg, text]]
        else:
            self.private_message_queue[user].append([msg, text])

    def private_message_queue_process(self, user):

        if user in self.private_message_queue:
            for data in self.private_message_queue[user][:]:
                msg, text = data
                self.private_message_queue[user].remove(data)
                self.privatechat.show_message(msg, text)

    def ip_ignored(self, address):

        if address is None:
            return True

        ips = self.config.sections["server"]["ipignorelist"]
        s_address = address.split(".")

        for ip in ips:

            # No Wildcard in IP
            if "*" not in ip:
                if address == ip:
                    return True
                continue

            # Wildcard in IP
            parts = ip.split(".")
            seg = 0

            for part in parts:
                # Stop if there's no wildcard or matching string number
                if part not in (s_address[seg], "*"):
                    break

                seg += 1

                # Last time around
                if seg == 4:
                    # Wildcard blocked
                    return True

        # Not blocked
        return False

    def say_chat_room(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            event = self.pluginhandler.incoming_public_chat_event(msg.room, msg.user, msg.msg)

            if event is not None:
                (r, n, msg.msg) = event
                self.chatrooms.say_chat_room(msg, msg.msg)
                self.pluginhandler.incoming_public_chat_notification(msg.room, msg.user, msg.msg)

    def add_user(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        self.watchedusers.add(msg.user)

        if not msg.userexists:
            if msg.user not in self.users:
                self.users[msg.user] = UserAddr(status=-1)

        if msg.status is not None:
            self.get_user_status(msg)
        elif msg.userexists and msg.status is None:
            self.queue.put(slskmessages.GetUserStatus(msg.user))

        if msg.files is not None:
            self.get_user_stats(msg)

    def privileged_users(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.set_privileged_users(msg.users)
            log.add(_("%i privileged users"), (len(msg.users)))
            self.queue.put(slskmessages.AddUser(self.config.sections["server"]["login"]))
            self.pluginhandler.server_connect_notification()

    def add_to_privileged(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.add_to_privileged(msg.user)

    def check_privileges(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        mins = msg.seconds // 60
        hours = mins // 60
        days = hours // 24

        if msg.seconds == 0:
            log.add(
                _("You have no privileges left. They are not necessary, but allow your downloads to be queued ahead of non-privileged users.")
            )
        else:
            log.add(
                _("%(days)i days, %(hours)i hours, %(minutes)i minutes, %(seconds)i seconds of download privileges left."), {
                    'days': days,
                    'hours': hours % 24,
                    'minutes': mins % 60,
                    'seconds': msg.seconds % 60
                }
            )

        self.privileges_left = msg.seconds

    def admin_message(self, msg):
        log.add("%s", (msg.msg))

    def child_depth(self, msg):
        # TODO: Implement me
        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

    def branch_level(self, msg):
        # TODO: Implement me
        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

    def branch_root(self, msg):
        # TODO: Implement me
        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

    def distrib_child_depth(self, msg):
        # TODO: Implement me
        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

    def distrib_branch_root(self, msg):
        # TODO: Implement me
        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

    def wishlist_interval(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.search is not None:
            self.search.wish_list.set_interval(msg)

    def get_user_status(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        # Causes recursive requests when privileged?
        # self.queue.put(slskmessages.AddUser(msg.user))
        if msg.user in self.users:
            if msg.status == 0:
                self.users[msg.user] = UserAddr(status=msg.status)
            else:
                self.users[msg.user].status = msg.status
        else:
            self.users[msg.user] = UserAddr(status=msg.status)

        if msg.privileged is not None:
            if msg.privileged == 1:
                if self.transfers is not None:
                    self.transfers.add_to_privileged(msg.user)
                else:
                    log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.interests is not None:
            self.interests.get_user_status(msg)

        if self.userlist is not None:
            self.userlist.get_user_status(msg)

        if self.transfers is not None:
            self.transfers.get_user_status(msg)

        if self.privatechat is not None:
            self.privatechat.get_user_status(msg)

        if self.userinfo is not None:
            self.userinfo.get_user_status(msg)

        if self.userbrowse is not None:
            self.userbrowse.get_user_status(msg)

        if self.chatrooms is not None:
            self.chatrooms.get_user_status(msg)

    def user_interests(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.userinfo is not None:
            self.userinfo.show_interests(msg)

    def get_user_stats(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if msg.user == self.config.sections["server"]["login"]:
            self.speed = msg.avgspeed

        if self.interests is not None:
            self.interests.get_user_stats(msg)

        if self.chatrooms is not None:
            self.chatrooms.get_user_stats(msg)

        if self.userinfo is not None:
            self.userinfo.get_user_stats(msg)

        if self.userlist is not None:
            self.userlist.get_user_stats(msg)

        stats = {
            'avgspeed': msg.avgspeed,
            'downloadnum': msg.downloadnum,
            'files': msg.files,
            'dirs': msg.dirs,
        }

        self.pluginhandler.user_stats_notification(msg.user, stats)

    def user_left_room(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.user_left_room(msg)

    def relogged(self, msg):
        log.add(_("Someone else is logging in with the same nickname, server is going to disconnect us"))
        self.manualdisconnect = True
        self.pluginhandler.server_disconnect_notification(False)

    def check_user(self, user, ip):
        """
        Check if this user is banned, geoip-blocked, and which shares
        it is allowed to access based on transfer and shares settings.
        """

        if user in self.config.sections["server"]["banlist"]:
            if self.config.sections["transfers"]["usecustomban"]:
                return 0, "Banned (%s)" % self.config.sections["transfers"]["customban"]
            else:
                return 0, "Banned"

        if user in (i[0] for i in self.config.sections["server"]["userlist"]):
            if self.config.sections["transfers"]["enablebuddyshares"]:
                # For sending buddy-only shares
                return 2, ""

            return 1, ""

        if self.config.sections["transfers"]["friendsonly"]:
            return 0, "Sorry, friends only"

        if ip is None or not self.config.sections["transfers"]["geoblock"]:
            return 1, ""

        cc = self.geoip.get_all(ip).country_short

        if cc == "-":
            if self.config.sections["transfers"]["geopanic"]:
                return 0, "Sorry, geographical paranoia"

            return 1, ""

        """ Please note that all country codes are stored in the same string at the first index
        of an array, separated by commas (no idea why...) """

        if self.config.sections["transfers"]["geoblockcc"][0].find(cc) >= 0:
            return 0, "Sorry, your country is blocked"

        return 1, ""

    def check_spoof(self, user, ip, port):

        if user not in self.users:
            return 0

        if self.users[user].addr is not None:

            if len(self.users[user].addr) == 2:
                if self.users[user].addr is not None:
                    u_ip, u_port = self.users[user].addr
                    if u_ip != ip:
                        log.add_warning(_("IP %(ip)s:%(port)s is spoofing user %(user)s with a peer request, blocking because it does not match IP: %(real_ip)s"), {
                            'ip': ip,
                            'port': port,
                            'user': user,
                            'real_ip': u_ip
                        })
                        return 1
        return 0

    def user_info_reply(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        conn = msg.conn.conn

        for i in self.peerconns:
            if i.conn is conn and self.userinfo is not None:
                # probably impossible to do this
                if i.username != self.config.sections["server"]["login"]:
                    indeterminate_progress = change_page = False
                    self.userinfo.show_user(i.username, msg.conn, msg, indeterminate_progress, change_page)
                    break

    def user_info_request(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        user = ip = port = None
        conn = msg.conn.conn

        # Get peer's username, ip and port
        for i in self.peerconns:
            if i.conn is conn:
                user = i.username
                if i.addr is not None:
                    ip, port = i.addr
                break

        if user is None:
            # No peer connection
            return

        request_time = time.time()

        if user in self.requested_info:
            if not request_time > 10 + self.requested_info[user]:
                # Ignoring request, because it's 10 or less seconds since the
                # last one by this user
                return

        self.requested_info[user] = request_time

        # Check address is spoofed, if possible
        if user == self.config.sections["server"]["login"]:

            if ip is not None and port is not None:
                log.add(
                    _("Blocking %(user)s from making a UserInfo request, possible spoofing attempt from IP %(ip)s port %(port)s"), {
                        'user': user,
                        'ip': ip,
                        'port': port
                    }
                )
            else:
                log.add(_("Blocking %s from making a UserInfo request, possible spoofing attempt from an unknown IP & port"), user)

            if conn is not None:
                self.queue.put(slskmessages.ConnClose(conn))

            return

        if user in self.config.sections["server"]["banlist"]:

            log.add_warning(
                _("%(user)s is banned, but is making a UserInfo request"), {
                    'user': user
                }
            )

            log.add_warning("%s %s", (msg.__class__, self.contents(msg)))

            return

        try:
            userpic = self.config.sections["userinfo"]["pic"]

            with open(userpic, 'rb') as f:
                pic = f.read()

        except Exception:
            pic = None

        descr = unescape(self.config.sections["userinfo"]["descr"])

        if self.transfers is not None:
            totalupl = self.transfers.get_total_uploads_allowed()
            queuesize = self.transfers.get_upload_queue_sizes()[0]
            slotsavail = self.transfers.allow_new_uploads()

            if self.config.sections["transfers"]["remotedownloads"]:
                uploadallowed = self.config.sections["transfers"]["uploadallowed"]
            else:
                uploadallowed = 0

            self.queue.put(slskmessages.UserInfoReply(conn, descr, pic, totalupl, queuesize, slotsavail, uploadallowed))

        log.add(
            _("%(user)s is making a UserInfo request"), {
                'user': user
            }
        )

    def shared_file_list(self, msg):

        conn = msg.conn.conn

        for i in self.peerconns:
            if i.conn is conn and self.userbrowse is not None:
                if i.username != self.config.sections["server"]["login"]:
                    indeterminate_progress = change_page = False
                    self.userbrowse.show_user(i.username, msg.conn, msg, indeterminate_progress, change_page)
                    break

    def file_search_result(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        conn = msg.conn
        addr = conn.addr

        if self.search is not None:
            if addr:
                country = self.geoip.get_all(addr[0]).country_short
            else:
                country = ""

            if country == "-":
                country = ""

            self.search.show_result(msg, msg.user, country)

            # Close peer connection immediately, otherwise we exhaust our connection limit
            self.close_peer_connection(conn, msg.user)

    def transfer_timeout(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.transfer_timeout(msg)

    def check_download_queue(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.check_download_queue()

    def file_download(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.file_download(msg)

    def file_upload(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.file_upload(msg)

    def file_request(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.file_request(msg)

    def file_error(self, msg):

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.file_error(msg)

    def transfer_request(self, msg):
        """ Peer code: 40 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.transfer_request(msg)

    def transfer_response(self, msg):
        """ Peer code: 41 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.transfer_response(msg)

    def queue_upload(self, msg):
        """ Peer code: 43 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.queue_upload(msg)

    def queue_failed(self, msg):
        """ Peer code: 50 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.queue_failed(msg)

    def place_in_queue_request(self, msg):
        """ Peer code: 51 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.place_in_queue_request(msg)

    def upload_queue_notification(self, msg):
        """ Peer code: 52 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.upload_queue_notification(msg)

    def upload_failed(self, msg):
        """ Peer code: 46 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.upload_failed(msg)

    def place_in_queue(self, msg):
        """ Peer code: 44 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.transfers is not None:
            self.transfers.place_in_queue(msg)

    def get_shared_file_list(self, msg):
        """ Peer code: 4 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        user = ip = port = None
        conn = msg.conn.conn

        # Get peer's username, ip and port
        for i in self.peerconns:
            if i.conn is conn:
                user = i.username
                if i.addr is not None:
                    if len(i.addr) != 2:
                        break
                    ip, port = i.addr
                break

        if user is None:
            # No peer connection
            return

        # Check address is spoofed, if possible
        # if self.check_spoof(user, ip, port):
        #     # Message IS spoofed
        #     return
        if user == self.config.sections["server"]["login"]:
            if ip is not None and port is not None:
                log.add(
                    _("%(user)s is making a BrowseShares request, blocking possible spoofing attempt from IP %(ip)s port %(port)s"), {
                        'user': user,
                        'ip': ip,
                        'port': port
                    })
            else:
                log.add(
                    _("%(user)s is making a BrowseShares request, blocking possible spoofing attempt from an unknown IP & port"), {
                        'user': user
                    })

            if conn is not None:
                self.queue.put(slskmessages.ConnClose(conn))
            return

        log.add(_("%(user)s is making a BrowseShares request"), {
            'user': user
        })

        ip, port = msg.conn.addr
        checkuser, reason = self.check_user(user, ip)

        if checkuser == 1:
            # Send Normal Shares
            if self.shares.newnormalshares:
                self.shares.create_compressed_shares_message("normal")
                self.shares.compress_shares("normal")
                self.shares.newnormalshares = False
            m = self.shares.compressed_shares_normal

        elif checkuser == 2:
            # Send Buddy Shares
            if self.shares.newbuddyshares:
                self.shares.create_compressed_shares_message("buddy")
                self.shares.compress_shares("buddy")
                self.shares.newbuddyshares = False
            m = self.shares.compressed_shares_buddy

        else:
            # Nyah, Nyah
            m = slskmessages.SharedFileList(conn, {})
            m.make_network_message(nozlib=0)

        m.conn = conn
        self.queue.put(m)

    def folder_contents_request(self, msg):
        """ Peer code: 36 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        conn = msg.conn.conn
        ip, port = msg.conn.addr
        username = None
        checkuser = None
        reason = ""

        for i in self.peerconns:
            if i.conn is conn:
                username = i.username
                checkuser, reason = self.check_user(username, ip)
                break

        if not username:
            return
        if not checkuser:
            self.queue.put(slskmessages.MessageUser(username, "[Automatic Message] " + reason))
            return

        if checkuser == 1:
            shares = self.shares.share_dbs["streams"]
        elif checkuser == 2:
            shares = self.shares.share_dbs["streams"]
        else:
            self.queue.put(slskmessages.TransferResponse(conn, 0, reason=reason, req=0))
            shares = {}

        if checkuser:
            if msg.dir in shares:
                self.queue.put(slskmessages.FolderContentsResponse(conn, msg.dir, shares[msg.dir]))
            elif msg.dir.rstrip('\\') in shares:
                self.queue.put(slskmessages.FolderContentsResponse(conn, msg.dir, shares[msg.dir.rstrip('\\')]))
            else:
                if checkuser == 2:
                    shares = self.shares.share_dbs["streams"]
                    if msg.dir in shares:
                        self.queue.put(slskmessages.FolderContentsResponse(conn, msg.dir, shares[msg.dir]))
                    elif msg.dir.rstrip("\\") in shares:
                        self.queue.put(slskmessages.FolderContentsResponse(conn, msg.dir, shares[msg.dir.rstrip("\\")]))

    def folder_contents_response(self, msg):
        """ Peer code: 37 """

        if self.transfers is not None:
            conn = msg.conn.conn
            file_list = msg.list

            # Check for a large number of files
            many = False
            folder = ""

            for i in file_list:
                for j in file_list[i]:
                    if os.path.commonprefix([i, j]) == j:
                        numfiles = len(file_list[i][j])
                        if numfiles > 100:
                            many = True
                            folder = j

            if many:
                for i in self.peerconns:
                    if i.conn is conn:
                        username = i.username
                        break

                self.transfers.downloadsview.download_large_folder(username, folder, numfiles, conn, file_list)
            else:
                self.transfers.folder_contents_response(conn, file_list)

    def room_list(self, msg):
        """ Server code: 64 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.set_room_list(msg)
            self.set_status("")

    def global_user_list(self, msg):
        """ Server code: 67 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.globallist is not None:
            self.globallist.set_global_users_list(msg)

    def peer_transfer(self, msg):
        if self.userinfo is not None and msg.msg is slskmessages.UserInfoReply:
            self.userinfo.update_gauge(msg)

        if self.userbrowse is not None and msg.msg is slskmessages.SharedFileList:
            self.userbrowse.update_gauge(msg)

    def tunneled_message(self, msg):
        """ Server code: 68 """
        """ DEPRECATED """

        if msg.code in self.protothread.peerclasses:
            peermsg = self.protothread.peerclasses[msg.code](None)
            peermsg.parse_network_message(msg.msg)
            peermsg.tunneleduser = msg.user
            peermsg.tunneledreq = msg.req
            peermsg.tunneledaddr = msg.addr
            self.network_callback([peermsg])
        else:
            log.add_msg_contents("Unknown tunneled message: %s", self.contents(msg))

    def file_search_request(self, msg):
        """ Peer code: 8 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        conn = msg.conn.conn

        for i in self.peerconns:
            if i.conn == conn:
                user = i.username
                self.shares.process_search_request(msg.searchterm, user, msg.searchid, direct=True)
                break

    def search_request(self, msg):
        """ Server code: 93 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        self.shares.process_search_request(msg.searchterm, msg.user, msg.searchid, direct=False)
        self.pluginhandler.search_request_notification(msg.searchterm, msg.user, msg.searchid)

    def room_search_request(self, msg):
        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))
        self.shares.process_search_request(msg.searchterm, msg.room, msg.searchid, direct=False)

    def distrib_search(self, msg):
        """ Distrib code: 3 """

        self.shares.process_search_request(msg.searchterm, msg.user, msg.searchid, direct=False)
        self.pluginhandler.distrib_search_notification(msg.searchterm, msg.user, msg.searchid)

    def possible_parents(self, msg):
        """ Server code: 102 """

        """ Server sent a list of 10 potential parents, whose purpose is to forward us search requests.
        We attempt to connect to them all at once, since connection errors are fairly common. """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        potential_parents = msg.list

        if not self.has_parent and potential_parents:

            for user in potential_parents:
                addr = potential_parents[user]

                self.send_message_to_peer(user, slskmessages.DistribConn(), address=addr)

    def get_parent_conn(self):

        for i in self.peerconns:
            if i.type == 'D':
                return i

        return None

    def parent_conn_closed(self):
        """ Tell the server it needs to send us a NetInfo message with a new list of
        potential parents. """

        self.has_parent = False
        self.queue.put(slskmessages.HaveNoParent(1))

    def distrib_branch_level(self, msg):
        """ Distrib code: 4 """

        """ This message is received when we have a successful connection with a potential
        parent. Tell the server who our parent is, and stop requesting new potential parents. """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if not self.has_parent:

            for i in self.peerconns[:]:
                if i.type == 'D':
                    """ We previously attempted to connect to all potential parents. Since we now
                    have a parent, stop connecting to the others. """

                    if i.conn != msg.conn.conn:
                        if i.conn is not None:
                            self.queue.put(slskmessages.ConnClose(i.conn, callback=False))

                        if i.conntimer is not None:
                            i.conntimer.cancel()

                        self.peerconns.remove(i)

            parent = self.get_parent_conn()

            if parent is not None:
                self.queue.put(slskmessages.HaveNoParent(0))
                self.queue.put(slskmessages.SearchParent(msg.conn.addr[0]))
                self.has_parent = True
            else:
                self.parent_conn_closed()

    def global_recommendations(self, msg):
        """ Server code: 56 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.interests is not None:
            self.interests.global_recommendations(msg)

    def recommendations(self, msg):
        """ Server code: 54 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.interests is not None:
            self.interests.recommendations(msg)

    def item_recommendations(self, msg):
        """ Server code: 111 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.interests is not None:
            self.interests.item_recommendations(msg)

    def similar_users(self, msg):
        """ Server code: 110 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.interests is not None:
            self.interests.similar_users(msg)

    def room_ticker_state(self, msg):
        """ Server code: 113 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.ticker_set(msg)

    def room_ticker_add(self, msg):
        """ Server code: 114 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.ticker_add(msg)

    def room_ticker_remove(self, msg):
        """ Server code: 115 """

        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

        if self.chatrooms is not None:
            self.chatrooms.ticker_remove(msg)

    def update_debug_log_options(self):
        """ Gives the logger updated logging settings """

        should_log = self.config.sections["logging"]["debug_file_output"]
        log_folder = self.config.sections["logging"]["debuglogsdir"]
        timestamp_format = self.config.sections["logging"]["log_timestamp"]

        log.set_log_to_file(should_log)
        log.set_folder(log_folder)
        log.set_timestamp_format(timestamp_format)

    def dummy_message(self, msg):
        log.add_msg_contents("%s %s", (msg.__class__, self.contents(msg)))

    def ignore(self, msg):
        # Ignore received message
        pass


class UserAddr:

    def __init__(self, addr=None, status=None):
        self.addr = addr
        self.status = status



from pynicotine.pluginsystem import BasePlugin


def enable(plugins):
    global PLUGIN
    PLUGIN = Plugin(plugins)


def disable(plugins):
    global PLUGIN
    PLUGIN = None


class Plugin(BasePlugin):
    __name__ = "Test Replier"

    def IncomingPublicChatEvent(self, room, nick, line):  # noqa
        if line.lower() == 'test':
            self.saypublic(room, 'Test failed.')



# COPYRIGHT (C) 2020-2021 Nicotine+ Team
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gi
import os
import sys

from gi.repository import Gdk
from gi.repository import GLib

from pynicotine import slskmessages
from pynicotine.gtkgui.dialogs import combo_box_dialog
from pynicotine.gtkgui.utils import PopupMenu
from pynicotine.logfacility import log


class Tray:

    def __init__(self, frame):

        try:
            # Check if AyatanaAppIndicator3 is available
            gi.require_version('AyatanaAppIndicator3', '0.1')
            from gi.repository import AyatanaAppIndicator3
            self.appindicator = AyatanaAppIndicator3

        except (ImportError, ValueError):
            try:
                # Check if AppIndicator3 is available
                gi.require_version('AppIndicator3', '0.1')
                from gi.repository import AppIndicator3
                self.appindicator = AppIndicator3

            except (ImportError, ValueError):
                # No AppIndicator support, fall back to GtkStatusIcon
                from gi.repository import Gtk
                self.appindicator = None
                self.gtk = Gtk

        self.frame = frame
        self.trayicon = None
        self.custom_icons = False
        self.local_icons = False
        self.final_icon_path = None
        self.tray_status = {
            "status": "disconnect",
            "last": "disconnect"
        }
        self.create_menu()

    def create_menu(self):

        try:
            self.tray_popup_menu_server = popup0 = PopupMenu(self, False)
            popup0.setup(
                ("#" + _("Connect"), self.frame.on_connect),
                ("#" + _("Disconnect"), self.frame.on_disconnect)
            )

            self.tray_popup_menu = popup = PopupMenu(self, False)
            popup.setup(
                ("#" + _("Hide / Show Nicotine+"), self.on_hide_unhide_window),
                (1, _("Server"), self.tray_popup_menu_server, None),
                ("#" + _("Downloads"), self.on_downloads),
                ("#" + _("Uploads"), self.on_uploads),
                ("#" + _("Send Message"), self.on_open_private_chat),
                ("#" + _("Lookup a User's IP"), self.on_get_a_users_ip),
                ("#" + _("Lookup a User's Info"), self.on_get_a_users_info),
                ("#" + _("Lookup a User's Shares"), self.on_get_a_users_shares),
                ("$" + _("Away"), self.frame.on_away),
                ("#" + _("Preferences"), self.frame.on_settings),
                ("#" + _("Quit"), self.frame.on_quit)
            )
        except Exception as e:
            log.add_warning(_('ERROR: tray menu, %(error)s'), {'error': e})

    def on_hide_unhide_window(self, widget):

        if self.frame.MainWindow.get_property("visible"):
            self.frame.MainWindow.hide()
        else:
            self.show_window()

    def on_downloads(self, widget):
        self.frame.on_downloads()
        self.show_window()

    def on_uploads(self, widget):
        self.frame.on_uploads()
        self.show_window()

    def on_open_private_chat(self, widget, prefix=""):

        # popup
        users = []
        for entry in self.frame.np.config.sections["server"]["userlist"]:
            users.append(entry[0])

        users.sort()
        user = combo_box_dialog(
            parent=self.frame.MainWindow,
            title=GLib.get_application_name() + ": " + _("Start Messaging"),
            message=_('Enter the User who you wish to send a private message:'),
            droplist=users
        )

        if user:
            self.frame.privatechats.send_message(user, show_user=True)
            self.frame.change_main_page("private")
            self.show_window()

    def on_get_a_users_info(self, widget, prefix=""):

        # popup
        users = []
        for entry in self.frame.np.config.sections["server"]["userlist"]:
            users.append(entry[0])

        users.sort()

        user = combo_box_dialog(
            parent=self.frame.MainWindow,
            title=GLib.get_application_name() + ": " + _("Get User Info"),
            message=_('Enter the User whose User Info you wish to receive:'),
            droplist=users
        )

        if user:
            self.frame.local_user_info_request(user)

    def on_get_a_users_ip(self, widget, prefix=""):
        users = []

        for entry in self.frame.np.config.sections["server"]["userlist"]:
            users.append(entry[0])

        users.sort()

        user = combo_box_dialog(
            parent=self.frame.MainWindow,
            title=GLib.get_application_name() + ": " + _("Get A User's IP"),
            message=_('Enter the User whose IP Address you wish to receive:'),
            droplist=users
        )

        if user:
            self.frame.np.ip_requested.add(user)
            self.frame.np.queue.put(slskmessages.GetPeerAddress(user))

    def on_get_a_users_shares(self, widget, prefix=""):

        users = []
        for entry in self.frame.np.config.sections["server"]["userlist"]:
            users.append(entry[0])

        users.sort()

        user = combo_box_dialog(
            parent=self.frame.MainWindow,
            title=GLib.get_application_name() + ": " + _("Get A User's Shares List"),
            message=_('Enter the User whose Shares List you wish to receive:'),
            droplist=users
        )

        if user:
            self.frame.browse_user(user)

    # GtkStatusIcon fallback
    def on_status_icon_popup(self, status_icon, button, activate_time):
        if button == 3:
            self.tray_popup_menu.popup(use_legacy=True)

    def check_icon_path(self, icon_name, icon_path, icon_type="local"):

        """
        Check if tray icons exist in the specified icon path.
        There are two naming schemes for tray icons:
        - System-wide/local icons: "org.nicotine_plus.Nicotine-<icon_name>"
        - Custom icons: "trayicon_<icon_name>"
        """

        if icon_type == "local":
            icon_scheme = GLib.get_prgname() + "-" + icon_name + "."
        else:
            icon_scheme = "trayicon_" + icon_name + "."

        try:
            scandir = os.scandir(icon_path)

            for entry in scandir:
                if entry.is_file() and entry.name.startswith(icon_scheme):
                    try:
                        scandir.close()
                    except AttributeError:
                        # Python 3.5 compatibility
                        pass

                    return True

        except FileNotFoundError:
            pass

        return False

    def get_final_icon_path(self):

        """ Returns an icon path to use for tray icons, or None to fall back to
        system-wide icons. """

        custom_icon_path = self.frame.np.config.sections["ui"]["icontheme"]

        if hasattr(sys, "real_prefix") or sys.base_prefix != sys.prefix:
            # Virtual environment
            local_icon_path = os.path.join(sys.prefix, "share", "icons", "hicolor", "scalable", "apps")
        else:
            # Git folder
            local_icon_path = os.path.abspath(os.path.join(self.frame.gui_dir, "..", "..", "files", "icons", "tray"))

        for icon_name in ("away", "connect", "disconnect", "msg"):

            # Check if custom icons exist
            if self.check_icon_path(icon_name, custom_icon_path, icon_type="custom"):
                self.custom_icons = True
                return custom_icon_path

            # Check if local icons exist
            if self.check_icon_path(icon_name, local_icon_path, icon_type="local"):
                self.local_icons = True
                return local_icon_path

        return None

    def load(self):

        """ Create """

        if self.trayicon is None:
            if self.appindicator is not None:
                trayicon = self.appindicator.Indicator.new(
                    GLib.get_application_name(),
                    "",
                    self.appindicator.IndicatorCategory.APPLICATION_STATUS)
                trayicon.set_menu(self.tray_popup_menu)

                # Action to hide/unhide main window when middle clicking the tray icon
                hide_unhide_item = self.tray_popup_menu.get_items()[_("Hide / Show Nicotine+")]
                trayicon.set_secondary_activate_target(hide_unhide_item)

            else:
                # GtkStatusIcon fallback
                trayicon = self.gtk.StatusIcon()
                trayicon.set_tooltip_text(GLib.get_application_name())
                trayicon.connect("activate", self.on_hide_unhide_window)
                trayicon.connect("popup-menu", self.on_status_icon_popup)

            self.trayicon = trayicon

        """ Set up icons """

        self.final_icon_path = self.get_final_icon_path()

        # If custom icon path was found, use it, otherwise we fall back to system icons
        if self.appindicator is not None and self.final_icon_path:
            self.trayicon.set_icon_theme_path(self.final_icon_path)

        """ Set visible """
        if self.appindicator is not None:
            self.trayicon.set_status(self.appindicator.IndicatorStatus.ACTIVE)
        else:
            # GtkStatusIcon fallback
            self.trayicon.set_visible(True)

        self.set_image(self.tray_status["status"])

    def hide(self):

        if not self.is_tray_icon_visible():
            return

        if self.appindicator is not None:
            self.trayicon.set_status(self.appindicator.IndicatorStatus.PASSIVE)
        else:
            # GtkStatusIcon fallback
            self.trayicon.set_visible(False)

    def show_window(self):
        self.frame.MainWindow.present_with_time(Gdk.CURRENT_TIME)

        self.frame.chatrooms.clear_notifications()
        self.frame.privatechats.clear_notifications()

    def is_tray_icon_visible(self):

        if self.trayicon is None:
            return False

        if self.appindicator is None:
            return self.trayicon.get_visible()

        if self.appindicator is not None and self.trayicon.get_status() != self.appindicator.IndicatorStatus.ACTIVE:
            return False

        return True

    def set_image(self, status=None):
        if not self.is_tray_icon_visible():
            return

        try:
            if status is not None:
                self.tray_status["status"] = status

            # Check for hilites, and display hilite icon if there is a room or private hilite
            if self.frame.hilites["rooms"] == [] and self.frame.hilites["private"] == []:
                # If there is no hilite, display the status
                icon_name = self.tray_status["status"]
            else:
                icon_name = "msg"

            if icon_name != self.tray_status["last"]:
                self.tray_status["last"] = icon_name

            if self.appindicator is not None:
                if self.custom_icons:
                    icon_name = "trayicon_" + icon_name
                else:
                    icon_name = GLib.get_prgname() + "-" + icon_name

                self.trayicon.set_icon_full(icon_name, GLib.get_application_name())

            else:
                # GtkStatusIcon fallback
                if self.custom_icons or self.local_icons:
                    self.trayicon.set_from_pixbuf(
                        self.frame.images["trayicon_" + icon_name]
                    )

                else:
                    self.trayicon.set_from_icon_name(GLib.get_prgname() + "-" + icon_name)

        except Exception as e:
            log.add_warning(_("ERROR: cannot set trayicon image: %(error)s"), {'error': e})

    def set_away(self, enable):

        if enable:
            self.tray_status["status"] = "away"
        else:
            self.tray_status["status"] = "connect"

        self.set_image()

        # Toggle away checkbox in tray menu
        away_item = self.tray_popup_menu.get_items()[_("Away")]
        handler_id = self.tray_popup_menu.handlers[away_item]

        with away_item.handler_block(handler_id):
            # Temporarily disable handler, we only want to change the visual checkbox appearance
            away_item.set_active(enable)

    def set_connected(self, enable):

        if enable:
            self.tray_status["status"] = "connect"
        else:
            self.tray_status["status"] = "disconnect"

        self.set_image()

    def set_server_actions_sensitive(self, status):

        items = self.tray_popup_menu.get_items()

        for i in (_("Send Message"), _("Lookup a User's IP"), _("Lookup a User's Info"),
                  _("Lookup a User's Shares"), _("Away")):

            """ Disable menu items when disconnected from server """
            items[i].set_sensitive(status)

        items = self.tray_popup_menu_server.get_items()

        items[_("Connect")].set_sensitive(not status)
        items[_("Disconnect")].set_sensitive(status)

    def set_transfer_status(self, download, upload):

        if self.trayicon is None:
            return

        items = self.tray_popup_menu.get_items()

        items[_("Downloads")].set_label(download)
        items[_("Uploads")].set_label(upload)



import gc

from pynicotine.pluginsystem import BasePlugin


def enable(plugins):
    global PLUGIN
    PLUGIN = Plugin(plugins)


def disable(plugins):
    global PLUGIN
    PLUGIN = None
    gc.set_debug(0)


class Plugin(BasePlugin):
    __name__ = "Memory Debugger"

    def init(self):
        self.log("""Tweaking garbage collection. Is it currently turned on? %s
Current thresholds: %s
Current counts: %s
Enabling GB debug output (check stderr)""" % (str(gc.isenabled()), repr(gc.get_threshold()), repr(gc.get_count())))

        gc.set_debug(gc.DEBUG_STATS)
        self.log("Forcing collection of generation 0...")
        gc.collect(0)
        self.log("Forcing collection of generation 1...")
        gc.collect(1)
        self.log("Forcing collection of generation 2...")
        gc.collect(2)
        unclaimed = ['A total of %s objects that could not be freed:' % (len(gc.garbage),)]
        for i in gc.garbage:
            unclaimed.append('%s: %s (%s)' % (type(i), str(i), repr(i)))
        self.log('\n'.join(unclaimed))
        self.log("Done.")
